import pathlib
import subprocess
from typing import Literal, Optional, Union, overload

_StrOrListStr = Union[str, list[str]]
_StrBytesPath = Union[str, bytes, pathlib.Path]


@overload
def run(
    command: _StrOrListStr,
    text: Literal[False] = False,
    check: bool = False,
    cwd: Optional[_StrBytesPath] = None,
    env: Optional[dict[str, str]] = None,
) -> subprocess.CompletedProcess[bytes]:
    ...


@overload
def run(
    command: _StrOrListStr,
    text: Literal[True],
    check: bool = False,
    cwd: Optional[_StrBytesPath] = None,
    env: Optional[dict[str, str]] = None,
) -> subprocess.CompletedProcess[str]:
    ...


def run(
    command: _StrOrListStr,
    text: bool = False,
    check: bool = False,
    cwd: Optional[_StrBytesPath] = None,
    env: Optional[dict[str, str]] = None,
) -> Union[subprocess.CompletedProcess[str], subprocess.CompletedProcess[bytes]]:
    return subprocess.run(
        command,
        shell=isinstance(command, str),
        text=text,
        check=check,
        env=env,
        cwd=cwd,
        capture_output=True,
    )



from .dependency import SourceCodeStat, VerificationStatus, resolve_dependency
from .file import (
    AddtionalSource,
    VerificationFile,
    VerificationInput,
    VerificationInputImpl,
)
from .result import FileResult, VerificationResult, VerifyCommandResult
from .result_status import ResultStatus
from .verification import (
    BaseVerification,
    CommandVerification,
    ConstVerification,
    ProblemVerification,
    Verification,
    VerificationParams,
)

__all__ = [
    "SourceCodeStat",
    "VerificationStatus",
    "resolve_dependency",
    "VerificationFile",
    "VerificationInput",
    "VerificationInputImpl",
    "AddtionalSource",
    "FileResult",
    "ResultStatus",
    "VerifyCommandResult",
    "BaseVerification",
    "CommandVerification",
    "ConstVerification",
    "ProblemVerification",
    "Verification",
    "VerificationResult",
    "VerificationParams",
]



# Python Version: 3.x
import functools
import pathlib
from logging import getLogger
from typing import Any, Optional

import competitive_verifier_oj_clone.shlex2 as shlex
from competitive_verifier_oj_clone.config import get_config
from competitive_verifier_oj_clone.languages.models import Language, LanguageEnvironment
from competitive_verifier_oj_clone.utils import read_text_normalized

logger = getLogger(__name__)


class NimLanguageEnvironment(LanguageEnvironment):
    compile_to: str
    NIMFLAGS: list[str]

    def __init__(self, *, compile_to: str, NIMFLAGS: list[str]):
        self.compile_to = compile_to
        self.NIMFLAGS = NIMFLAGS  # type: ignore

    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        return shlex.join(
            [
                "nim",
                self.compile_to,
                "-p:.",
                f"-o:{str(tempdir /'a.out')}",
                f"--nimcache:{str(tempdir)}",
            ]
            + self.NIMFLAGS
            + [str(path)]
        )

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        return str(tempdir / "a.out")


@functools.lru_cache(maxsize=None)
def _list_direct_dependencies(
    path: pathlib.Path, *, basedir: pathlib.Path
) -> list[pathlib.Path]:
    items: list[str] = []
    for line in read_text_normalized(basedir / path).splitlines():
        line = line.strip()
        if line.startswith("include"):
            items += line[7:].strip().split(",")
        elif line.startswith("import"):
            line = line[6:]
            i = line.find(" except ")
            if i >= 0:
                line = line[:i]
            items += line.split(",")
        elif line.startswith("from"):
            i = line.find(" import ")
            if i >= 0:
                items += line[4 : i - 1]
    dependencies = [path.resolve()]
    for item in items:
        item = item.strip()
        if item.startswith('"'):
            item = item[1 : len(item) - 1]
        else:
            item += ".nim"
        item_ = pathlib.Path(item)
        if item_.exists():
            dependencies.append(item_)
    return list(set(dependencies))


class NimLanguage(Language):
    config: dict[str, Any]

    def __init__(self, *, config: Optional[dict[str, Any]] = None):
        if config is None:
            self.config = get_config()["languages"].get("nim", {})
        else:
            self.config = config

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        dependencies: list[pathlib.Path] = []
        visited: set[pathlib.Path] = set()
        stk = [path.resolve()]
        while stk:
            path = stk.pop()
            if path in visited:
                continue
            visited.add(path)
            for child in _list_direct_dependencies(path, basedir=basedir):
                dependencies.append(child)
                stk.append(child)
        return list(set(dependencies))

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[NimLanguageEnvironment]:
        default_compile_to = "cpp"
        default_NIMFLAGS = ["-d:release", "--opt:speed"]
        envs: list[NimLanguageEnvironment] = []
        if "environments" not in self.config:
            envs.append(
                NimLanguageEnvironment(
                    compile_to=default_compile_to, NIMFLAGS=default_NIMFLAGS
                )
            )
        else:
            for env in self.config["environments"]:
                compile_to = env.get("compile_to", default_compile_to)
                NIMFLAGS: list[str] = env.get("NIMFLAGS", default_NIMFLAGS)
                if not isinstance(NIMFLAGS, list):  # type:ignore
                    raise RuntimeError("NIMFLAGS must ba a list")
                envs.append(
                    NimLanguageEnvironment(compile_to=compile_to, NIMFLAGS=NIMFLAGS)
                )
        return envs



from typing import Any

import pytest

from competitive_verifier.verify.split_state import SplitState

test_split_by_split_state_params: list[tuple[SplitState, list[Any], list[Any]]] = [
    (SplitState(size=3, index=0), [0, 1, 2, 3, 4], [0]),
    (SplitState(size=3, index=1), [0, 1, 2, 3, 4], [1, 2]),
    (SplitState(size=3, index=2), [0, 1, 2, 3, 4], [3, 4]),
    (SplitState(size=6, index=4), [0, 1, 2, 3, 4], [4]),
    (SplitState(size=6, index=5), [0, 1, 2, 3, 4], []),
    (SplitState(size=3, index=2), ["a", "b", "c", "d", "e"], ["d", "e"]),
]


@pytest.mark.parametrize("state, lst, expected", test_split_by_split_state_params)
def test_split_by_split_state(state: SplitState, lst: list[Any], expected: list[Any]):
    assert state.split(lst) == expected



import argparse
import math
import os
import pathlib
from typing import Any

import pytest

import competitive_verifier.console.app as app
from competitive_verifier.arg import COMPETITIVE_VERIFY_FILES_PATH


def parse_args(args: list[str]) -> argparse.Namespace:
    return app.get_parser().parse_args(args)


@pytest.fixture
def setenv():
    os.environ[
        COMPETITIVE_VERIFY_FILES_PATH
    ] = ".competitive-verifier/verify_files.json"
    yield
    del os.environ[COMPETITIVE_VERIFY_FILES_PATH]


def test_parse_args_default(setenv: Any):
    parsed = parse_args(["verify"])

    assert parsed.verify_files_json == pathlib.Path(
        ".competitive-verifier/verify_files.json"
    )
    assert parsed.timeout == math.inf
    assert parsed.default_tle == 60.0
    assert parsed.prev_result is None
    assert parsed.split is None
    assert parsed.split_index is None

    parsed = parse_args(
        [
            "docs",
            ".cr/res0.json",
            ".cr/res1.json",
        ]
    )

    assert parsed.verify_files_json == pathlib.Path(
        ".competitive-verifier/verify_files.json"
    )
    assert parsed.result_json == [
        pathlib.Path(".cr/res0.json"),
        pathlib.Path(".cr/res1.json"),
    ]


def test_parse_args_json_path(setenv: Any):
    parsed = parse_args(["verify", "--verify-json", ".cv/f.json"])
    assert parsed.verify_files_json == pathlib.Path(".cv/f.json")

    parsed = parse_args(
        [
            "docs",
            "--verify-json",
            ".cv/d.json",
            ".cr/res0.json",
            ".cr/res1.json",
        ]
    )
    assert parsed.verify_files_json == pathlib.Path(".cv/d.json")
    assert parsed.result_json == [
        pathlib.Path(".cr/res0.json"),
        pathlib.Path(".cr/res1.json"),
    ]


def test_parse_args_time(setenv: Any):
    parsed = parse_args(["verify", "--timeout", "600", "--tle", "10"])

    assert parsed.timeout == 600.0
    assert parsed.default_tle == 10.0
    assert parsed.prev_result is None


def test_parse_args_prev_result(setenv: Any):
    parsed = parse_args(["verify", "--prev-result", ".cv/prev.json"])
    assert parsed.prev_result == pathlib.Path(".cv/prev.json")


def test_parallel_split(setenv: Any):
    parsed = parse_args(["verify", "--split-index", "1", "--split", "5"])
    assert parsed.split == 5
    assert parsed.split_index == 1



import pathlib
from typing import Any

import pytest

from competitive_verifier.models import (
    AddtionalSource,
    CommandVerification,
    ConstVerification,
    ResultStatus,
    VerificationFile,
)

test_parse_VerificationFile_params: list[
    tuple[VerificationFile, dict[str, Any], dict[str, Any]]
] = [
    (
        VerificationFile(),
        {},
        {
            "dependencies": set(),
            "document_attributes": {},
            "verification": [],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            dependencies=set(
                [
                    pathlib.Path("bar1"),
                    pathlib.Path("bar2"),
                ]
            ),
        ),
        {
            "dependencies": [
                "bar1",
                "bar2",
            ],
        },
        {
            "dependencies": set(
                [
                    pathlib.Path("bar1"),
                    pathlib.Path("bar2"),
                ]
            ),
            "document_attributes": {},
            "verification": [],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            document_attributes={
                "title": "Bar bar",
            },
        ),
        {
            "document_attributes": {
                "title": "Bar bar",
            },
        },
        {
            "dependencies": set(),
            "document_attributes": {
                "title": "Bar bar",
            },
            "verification": [],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            verification=[ConstVerification(status=ResultStatus.SUCCESS)],
        ),
        {
            "verification": [
                {
                    "type": "const",
                    "status": "success",
                }
            ],
        },
        {
            "dependencies": set(),
            "document_attributes": {},
            "verification": [ConstVerification(status=ResultStatus.SUCCESS)],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            verification=[ConstVerification(status=ResultStatus.SUCCESS)],
        ),
        {
            "verification": {
                "type": "const",
                "status": "success",
            },
        },
        {
            "dependencies": set(),
            "document_attributes": {},
            "verification": [ConstVerification(status=ResultStatus.SUCCESS)],
            "additonal_sources": [],
        },
    ),
    (
        VerificationFile(
            additonal_sources=[
                AddtionalSource(name="dummy", path=pathlib.Path("tmp/dummy.sh"))
            ]
        ),
        {
            "additonal_sources": [{"name": "dummy", "path": "tmp/dummy.sh"}],
        },
        {
            "dependencies": set(),
            "document_attributes": {},
            "verification": [],
            "additonal_sources": [
                {"name": "dummy", "path": pathlib.Path("tmp/dummy.sh")}
            ],
        },
    ),
]


@pytest.mark.parametrize(
    "obj, raw_dict, output_dict",
    test_parse_VerificationFile_params,
)
def test_parse_VerificationFile(
    obj: VerificationFile,
    raw_dict: dict[str, Any],
    output_dict: dict[str, Any],
):
    assert obj == VerificationFile.parse_obj(raw_dict)
    assert obj.dict() == output_dict


test_is_verification_params = [
    (
        VerificationFile(
            verification=[ConstVerification(status=ResultStatus.SUCCESS)],
        ),
        True,
        True,
    ),
    (
        VerificationFile(
            verification=[
                ConstVerification(status=ResultStatus.SUCCESS),
                ConstVerification(status=ResultStatus.FAILURE),
            ],
        ),
        True,
        True,
    ),
    (
        VerificationFile(
            verification=[CommandVerification(command="true")],
        ),
        True,
        False,
    ),
    (
        VerificationFile(
            verification=[
                ConstVerification(status=ResultStatus.SUCCESS),
                CommandVerification(command="true"),
            ],
        ),
        True,
        False,
    ),
    (
        VerificationFile(
            verification=[],
        ),
        False,
        False,
    ),
]


@pytest.mark.parametrize(
    "obj, is_verification, is_skippable_verification", test_is_verification_params
)
def test_is_verification(
    obj: VerificationFile,
    is_verification: bool,
    is_skippable_verification: bool,
):
    assert obj.is_verification() == is_verification
    assert obj.is_skippable_verification() == is_skippable_verification



import subprocess
import sys
from contextlib import nullcontext
from logging import getLogger
from typing import Literal, Optional, Union, overload

from competitive_verifier import log

logger = getLogger(__name__)

_StrOrListStr = Union[str, list[str]]


@overload
def exec_command(
    command: _StrOrListStr,
    text: Literal[False] = False,
    check: bool = False,
    capture_output: bool = False,
    env: Optional[dict[str, str]] = None,
    group_log: bool = False,
) -> subprocess.CompletedProcess[bytes]:
    ...


@overload
def exec_command(
    command: _StrOrListStr,
    text: Literal[True],
    check: bool = False,
    capture_output: bool = False,
    env: Optional[dict[str, str]] = None,
    group_log: bool = False,
) -> subprocess.CompletedProcess[str]:
    ...


def exec_command(
    command: _StrOrListStr,
    text: bool = False,
    check: bool = False,
    capture_output: bool = False,
    env: Optional[dict[str, str]] = None,
    group_log: bool = False,
) -> Union[subprocess.CompletedProcess[str], subprocess.CompletedProcess[bytes]]:
    if group_log:
        cm = log.group(f"subprocess.run: {command}")
    else:
        logger.info("subprocess.run: %s", command)
        cm = nullcontext()

    encoding = sys.stdout.encoding if text else None

    with cm:
        return subprocess.run(
            command,
            shell=isinstance(command, str),
            text=text,
            check=check,
            env=env,
            capture_output=capture_output,
            encoding=encoding,
        )



from typing import Any, Optional

from competitive_verifier_oj_clone.config import get_config
from competitive_verifier_oj_clone.languages.user_defined import UserDefinedLanguage


class RubyLanguage(UserDefinedLanguage):
    config: dict[str, Any]

    def __init__(self, *, config: Optional[dict[str, Any]] = None):
        if config is None:
            config = get_config()["languages"].get("ruby", {})
        assert config is not None
        config.setdefault("execute", "ruby {basedir}/{path}")
        super().__init__(extension="rb", config=config)



import sys
from contextlib import contextmanager
from logging import (
    CRITICAL,
    DEBUG,
    ERROR,
    NOTSET,
    WARNING,
    Handler,
    LogRecord,
    basicConfig,
)
from typing import Optional, TextIO

import colorlog
from colorama import Fore, Style

import competitive_verifier.github as github

_orig_stderr = sys.stderr


def override_stderr() -> None:
    sys.stderr = sys.stdout


def restore_stderr() -> None:
    sys.stderr = _orig_stderr


class GitHubActionsHandler(Handler):
    def __init__(self, *, stream: Optional[TextIO] = None) -> None:
        super().__init__(DEBUG)
        self.stream = stream

    def emit(self, record: LogRecord) -> None:
        message = record.getMessage()
        # file = record.pathname
        # line = record.levelno

        if record.levelno == DEBUG:
            github.print_debug(message, stream=self.stream)
        elif record.levelno == WARNING:
            github.print_warning(message, stream=self.stream)
        elif record.levelno == ERROR or record.levelno == CRITICAL:
            github.print_error(message, stream=self.stream)


# class ExceptGitHubActionsFilter(Filter):
#     def __init__(self) -> None:
#         super().__init__()

#     def filter(self, record: LogRecord) -> bool:
#         return (
#             record.levelno != WARNING
#             and record.levelno != ERROR
#             and record.levelno != CRITICAL
#         )


def configure_logging(
    default_level: Optional[int] = None,
    in_github_actions: Optional[bool] = None,
) -> None:
    if in_github_actions is None:
        in_github_actions = github.env.is_in_github_actions()

    override_stderr()

    colorlog_handler = colorlog.StreamHandler()
    colorlog_handler.setLevel(default_level or WARNING)
    colorlog_handler.setFormatter(
        colorlog.ColoredFormatter(
            "%(log_color)s%(levelname)s%(reset)s:%(name)s:%(message)s"
        )
    )
    handlers: list[Handler] = [colorlog_handler]

    if in_github_actions:
        handlers.append(GitHubActionsHandler())
        # colorlog_handler.addFilter(ExceptGitHubActionsFilter())

    basicConfig(
        level=NOTSET,
        handlers=handlers,
    )


def configure_stderr_logging(default_level: Optional[int] = None) -> None:
    colorlog_handler = colorlog.StreamHandler(sys.stderr)
    colorlog_handler.setLevel(default_level or WARNING)
    colorlog_handler.setFormatter(
        colorlog.ColoredFormatter(
            "%(log_color)s%(levelname)s%(reset)s:%(name)s:%(message)s"
        )
    )
    handlers: list[Handler] = [colorlog_handler]

    basicConfig(
        level=NOTSET,
        handlers=handlers,
    )


def _console_group(category: str, *, title: str, file: Optional[TextIO]):
    print(
        (
            f"<------------- {Fore.CYAN}{category}:"
            f"{Fore.YELLOW}{title}"
            f"{Style.RESET_ALL} ------------->"
        ),
        file=file,
    )


@contextmanager
def group(title: str, *, stream: Optional[TextIO] = None):
    if github.env.is_in_github_actions():
        try:
            github.begin_group(title, stream=stream)
            yield
        finally:
            github.end_group(stream=stream)
    else:
        try:
            _console_group(" Start group", title=title, file=stream)
            yield
        finally:
            _console_group("Finish group", title=title, file=stream)



# Python Version: 3.x
import functools
import json
import os
import pathlib
import re
import shutil
from logging import getLogger
from typing import Any, Optional

from .. import subprocess2 as subprocess

logger = getLogger(__name__)

BITS_STDCXX_H = "bits/stdc++.h"

CXX_STANDARD_LIBS = {
    "algorithm",
    "array",
    "bitset",
    "chrono",
    "codecvt",
    "complex",
    "condition_variable",
    "deque",
    "exception",
    "forward_list",
    "fstream",
    "functional",
    "future",
    "iomanip",
    "ios",
    "iosfwd",
    "iostream",
    "istream",
    "iterator",
    "limits",
    "list",
    "locale",
    "map",
    "memory",
    "mutex",
    "new",
    "numeric",
    "ostream",
    "queue",
    "random",
    "regex",
    "set",
    "sstream",
    "stack",
    "stdexcept",
    "streambuf",
    "string",
    "thread",
    "tuple",
    "typeinfo",
    "unordered_map",
    "unordered_set",
    "utility",
    "valarray",
    "vector",
}

C_STANDARD_LIBS = {
    "assert.h",
    "complex.h",
    "ctype.h",
    "errno.h",
    "fenv.h",
    "float.h",
    "inttypes.h",
    "iso646.h",
    "limits.h",
    "locale.h",
    "math.h",
    "setjmp.h",
    "signal.h",
    "stdalign.h",
    "stdarg.h",
    "stdatomic.h",
    "stdbool.h",
    "stddef.h",
    "stdint.h",
    "stdio.h",
    "stdlib.h",
    "stdnoreturn.h",
    "string.h",
    "tgmath.h",
    "threads.h",
    "time.h",
    "uchar.h",
    "wchar.h",
    "wctype.h",
}

CXX_C_ORIGIN_LIBS = {"c" + name[: -len(".h")] for name in C_STANDARD_LIBS}

BITS_EXTCXX_H = "bits/extc++.h"

EXT_LIBS = {
    "ext/algorithm",
    "ext/array_allocator.h",
    "ext/atomicity.h",
    "ext/bitmap_allocator.h",
    "ext/cast.h",
    "ext/concurrence.h",
    "ext/debug_allocator.h",
    "ext/extptr_allocator.h",
    "ext/functional",
    "ext/iterator",
    "ext/malloc_allocator.h",
    "ext/memory",
    "ext/mt_allocator.h",
    "ext/new_allocator.h",
    "ext/numeric",
    "ext/pod_char_traits.h",
    "ext/pointer.h",
    "ext/pool_allocator.h",
    "ext/rb_tree",
    "ext/rope",
    "ext/slist",
    "ext/stdio_filebuf.h",
    "ext/stdio_sync_filebuf.h",
    "ext/throw_allocator.h",
    "ext/typelist.h",
    "ext/type_traits.h",
    "ext/vstring.h",
    "ext/pb_ds/assoc_container.hpp",
    "ext/pb_ds/priority_queue.hpp",
    "ext/pb_ds/exception.hpp",
    "ext/pb_ds/hash_policy.hpp",
    "ext/pb_ds/list_update_policy.hpp",
    "ext/pb_ds/tree_policy.hpp",
    "ext/pb_ds/trie_policy.hpp",
}

BITS_STDTR1CXX_H = "bits/stdtr1c++.h"

TR1_LIBS = {
    "tr1/array",
    "tr1/cctype",
    "tr1/cfenv",
    "tr1/cfloat",
    "tr1/cinttypes",
    "tr1/climits",
    "tr1/cmath",
    "tr1/complex",
    "tr1/cstdarg",
    "tr1/cstdbool",
    "tr1/cstdint",
    "tr1/cstdio",
    "tr1/cstdlib",
    "tr1/ctgmath",
    "tr1/ctime",
    "tr1/cwchar",
    "tr1/cwctype",
    "tr1/functional",
    "tr1/random",
    "tr1/tuple",
    "tr1/unordered_map",
    "tr1/unordered_set",
    "tr1/utility",
}


@functools.lru_cache(maxsize=None)
def _check_compiler(compiler: str) -> str:
    # Executables named "g++" are not always g++, due to the fake g++ of macOS
    version = subprocess.run([compiler, "--version"], text=False).stdout.decode()
    if "clang" in version.lower() or "Apple LLVM".lower() in version.lower():
        return "clang"
    if "g++" in version.lower():
        return "gcc"
    return "unknown"  # default


@functools.lru_cache(maxsize=None)
def _get_uncommented_code(
    path: pathlib.Path, *, iquotes_options: tuple[str, ...], compiler: str
) -> bytes:
    # `iquotes_options` must be a tuple to use `lru_cache`

    if shutil.which(compiler) is None:
        raise BundleError(f"command not found: {compiler}")
    if _check_compiler(compiler) != "gcc":
        if compiler == "g++":
            raise BundleError(
                f"A fake g++ is detected. Please install the GNU C++ compiler.: {compiler}"
            )
        raise BundleError(
            f"It's not g++. Please specify g++ with $CXX envvar.: {compiler}"
        )
    command = [
        compiler,
        "-x",
        "c++",
        *iquotes_options,
        "-fpreprocessed",
        "-dD",
        "-E",
        str(path),
    ]
    return subprocess.run(command, text=False).stdout


def get_uncommented_code(
    path: pathlib.Path, *, iquotes: list[pathlib.Path], compiler: str
) -> bytes:
    iquotes_options: list[str] = []
    for iquote in iquotes:
        iquotes_options.extend(["-I", str(iquote.resolve())])
    code = _get_uncommented_code(
        path.resolve(), iquotes_options=tuple(iquotes_options), compiler=compiler
    )
    lines: list[bytes] = []
    for line in code.splitlines(keepends=True):
        m = re.match(rb'# (\d+) ".*"', line.rstrip())
        if m:
            lineno = int(m.group(1))
            while len(lines) + 1 < lineno:
                lines.append(b"\n")
        else:
            lines.append(line)
    return b"".join(lines)


class BundleError(Exception):
    pass


class BundleErrorAt(BundleError):
    def __init__(
        self, path: pathlib.Path, line: int, message: str, *args: Any, **kwargs: Any
    ):
        try:
            path = path.resolve().relative_to(pathlib.Path.cwd())
        except ValueError:
            pass
        message = "{}: line {}: {}".format(str(path), line, message)
        super().__init__(message, *args, **kwargs)  # type: ignore


class Bundler:
    iquotes: list[pathlib.Path]
    pragma_once: set[pathlib.Path]
    pragma_once_system: set[str]
    result_lines: list[bytes]
    path_stack: set[pathlib.Path]
    compiler: str

    def __init__(
        self,
        *,
        iquotes: Optional[list[pathlib.Path]] = None,
        compiler: str = os.environ.get("CXX", "g++"),
    ) -> None:
        if iquotes is None:
            iquotes = []
        self.iquotes = iquotes
        self.pragma_once = set()
        self.pragma_once_system = set()
        self.result_lines = []
        self.path_stack = set()
        self.compiler = compiler

    # これをしないと __FILE__ や __LINE__ が壊れる
    def _line(self, line: int, path: pathlib.Path) -> None:
        while self.result_lines and self.result_lines[-1].startswith(b"#line "):
            self.result_lines.pop()
        try:
            path = path.relative_to(pathlib.Path.cwd())
        except ValueError:
            pass
        # パス中の特殊文字を JSON style にエスケープしてから生成コードに記述
        # quick solution to this: https://github.com/online-judge-tools/verification-helper/issues/280
        self.result_lines.append(
            "#line {} {}\n".format(line, json.dumps(str(path))).encode()
        )

    # path を解決する
    # see: https://gcc.gnu.org/onlinedocs/gcc/Directory-Options.html#Directory-Options
    def _resolve(
        self, path: pathlib.Path, *, included_from: pathlib.Path
    ) -> pathlib.Path:
        if (included_from.parent / path).exists():
            return (included_from.parent / path).resolve()
        for dir_ in self.iquotes:
            if (dir_ / path).exists():
                return (dir_ / path).resolve()
        raise BundleErrorAt(path, -1, "no such header")

    def update(self, path: pathlib.Path) -> None:
        if path.resolve() in self.pragma_once:
            logger.debug(
                "%s: skipped since this file is included once with include guard",
                str(path),
            )
            return

        # 再帰的に自分自身を #include してたら諦める
        if path in self.path_stack:
            raise BundleErrorAt(path, -1, "cycle found in inclusion relations")
        self.path_stack.add(path)
        try:
            code = path.read_bytes()
            if not code.endswith(b"\n"):
                # ファイルの末尾に改行がなかったら足す
                code += b"\n"

            # include guard のまわりの変数
            # NOTE: include guard に使われたマクロがそれ以外の用途にも使われたり #undef されたりすると壊れるけど、無視します
            non_guard_line_found = False
            pragma_once_found = False
            include_guard_macro = None
            include_guard_define_found = False
            include_guard_endif_found = False
            preprocess_if_nest = 0

            lines = code.splitlines(keepends=True)
            uncommented_lines = get_uncommented_code(
                path, iquotes=self.iquotes, compiler=self.compiler
            ).splitlines(keepends=True)
            uncommented_lines.extend(
                [b""] * (len(lines) - len(uncommented_lines))
            )  # trailing comment lines are removed
            assert len(lines) == len(uncommented_lines)
            self._line(1, path)
            for i, (line, uncommented_line) in enumerate(zip(lines, uncommented_lines)):

                # nest の処理
                if re.match(rb"\s*#\s*(if|ifdef|ifndef)\s.*", uncommented_line):
                    preprocess_if_nest += 1
                if re.match(rb"\s*#\s*(else\s*|elif\s.*)", uncommented_line):
                    if preprocess_if_nest == 0:
                        raise BundleErrorAt(path, i + 1, "unmatched #else / #elif")
                if re.match(rb"\s*#\s*endif\s*", uncommented_line):
                    preprocess_if_nest -= 1
                    if preprocess_if_nest < 0:
                        raise BundleErrorAt(path, i + 1, "unmatched #endif")
                is_toplevel = preprocess_if_nest == 0 or (
                    preprocess_if_nest == 1 and include_guard_macro is not None
                )

                # #pragma once
                if re.match(
                    rb"\s*#\s*pragma\s+once\s*", line
                ):  # #pragma once は comment 扱いで消されてしまう
                    logger.debug("%s: line %s: #pragma once", str(path), i + 1)
                    if non_guard_line_found:
                        # 先頭以外で #pragma once されてた場合は諦める
                        raise BundleErrorAt(
                            path, i + 1, "#pragma once found in a non-first line"
                        )
                    if include_guard_macro is not None:
                        raise BundleErrorAt(
                            path,
                            i + 1,
                            "#pragma once found in an include guard with #ifndef",
                        )
                    if path.resolve() in self.pragma_once:
                        return
                    pragma_once_found = True
                    self.pragma_once.add(path.resolve())
                    self._line(i + 2, path)
                    continue

                # #ifndef HOGE_H as guard
                if (
                    not pragma_once_found
                    and not non_guard_line_found
                    and include_guard_macro is None
                ):
                    matched = re.match(rb"\s*#\s*ifndef\s+(\w+)\s*", uncommented_line)
                    if matched:
                        include_guard_macro = matched.group(1).decode()
                        logger.debug(
                            "%s: line %s: #ifndef %s",
                            str(path),
                            i + 1,
                            include_guard_macro,
                        )
                        self.result_lines.append(b"\n")
                        continue

                # #define HOGE_H as guard
                if include_guard_macro is not None and not include_guard_define_found:
                    matched = re.match(rb"\s*#\s*define\s+(\w+)\s*", uncommented_line)
                    if matched and matched.group(1).decode() == include_guard_macro:
                        self.pragma_once.add(path.resolve())
                        logger.debug(
                            "%s: line %s: #define %s",
                            str(path),
                            i + 1,
                            include_guard_macro,
                        )
                        include_guard_define_found = True
                        self.result_lines.append(b"\n")
                        continue

                # #endif as guard
                if (
                    include_guard_define_found
                    and preprocess_if_nest == 0
                    and not include_guard_endif_found
                ):
                    if re.match(rb"\s*#\s*endif\s*", uncommented_line):
                        include_guard_endif_found = True
                        self.result_lines.append(b"\n")
                        continue

                if uncommented_line and not re.match(rb"^\s*$", uncommented_line):
                    non_guard_line_found = True
                    if (
                        include_guard_macro is not None
                        and not include_guard_define_found
                    ):
                        # 先頭に #ifndef が見付かっても #define が続かないならそれは include guard ではない
                        include_guard_macro = None
                    if include_guard_endif_found:
                        # include guard の外側にコードが書かれているとまずいので検出する
                        raise BundleErrorAt(
                            path, i + 1, "found codes out of include guard"
                        )

                # #include <...>
                matched = re.match(rb"\s*#\s*include\s*<(.*)>\s*", uncommented_line)
                if matched:
                    included = matched.group(1).decode()
                    logger.debug(
                        "%s: line %s: #include <%s>", str(path), i + 1, str(included)
                    )
                    if included in self.pragma_once_system:
                        self._line(i + 2, path)
                    elif not is_toplevel:
                        # #pragma once 系の判断ができない場合はそっとしておく
                        self.result_lines.append(line)
                    elif (
                        included in C_STANDARD_LIBS
                        or included in CXX_STANDARD_LIBS
                        or included in CXX_C_ORIGIN_LIBS
                    ):
                        if BITS_STDCXX_H in self.pragma_once_system:
                            self._line(i + 2, path)
                        else:
                            self.pragma_once_system.add(included)
                            self.result_lines.append(line)
                    elif included in EXT_LIBS:
                        if BITS_EXTCXX_H in self.pragma_once_system:
                            self._line(i + 2, path)
                        else:
                            self.pragma_once_system.add(included)
                            self.result_lines.append(line)
                    elif included in TR1_LIBS:
                        if BITS_STDTR1CXX_H in self.pragma_once_system:
                            self._line(i + 2, path)
                        else:
                            self.pragma_once_system.add(included)
                            self.result_lines.append(line)
                    else:
                        # possibly: bits/*, tr2/* boost/*, c-posix library, etc.
                        self.pragma_once_system.add(included)
                        self.result_lines.append(line)
                        if included in [BITS_EXTCXX_H, BITS_STDTR1CXX_H]:
                            self.pragma_once_system.add(BITS_STDCXX_H)
                    continue

                # #include "..."
                matched = re.match(rb'\s*#\s*include\s*"(.*)"\s*', uncommented_line)
                if matched:
                    included = matched.group(1).decode()
                    logger.debug(
                        '%s: line %s: #include "%s"', str(path), i + 1, included
                    )
                    if not is_toplevel:
                        # #if の中から #include されると #pragma once 系の判断が不可能になるので諦める
                        raise BundleErrorAt(
                            path,
                            i + 1,
                            "unable to process #include in #if / #ifdef / #ifndef other than include guards",
                        )
                    self.update(
                        self._resolve(pathlib.Path(included), included_from=path)
                    )
                    self._line(i + 2, path)
                    # #include "iostream" みたいに書いたときの挙動をはっきりさせる
                    # #include <iostream> /* とかをやられた場合を落とす
                    continue

                # otherwise
                self.result_lines.append(line)

            # #if #endif の対応が壊れてたら諦める
            last_index = i + 1  # pyright: reportUnboundVariable=false
            if preprocess_if_nest != 0:
                raise BundleErrorAt(
                    path, last_index, "unmatched #if / #ifdef / #ifndef"
                )
            if include_guard_macro is not None and not include_guard_endif_found:
                raise BundleErrorAt(path, last_index, "unmatched #ifndef")

        finally:
            # 中で return することがあるので finally 節に入れておく
            self.path_stack.remove(path)

    def get(self) -> bytes:
        return b"".join(self.result_lines)



import datetime
import pathlib
from logging import getLogger
from typing import Any, Union

from pydantic import BaseModel, Field, validator

from competitive_verifier.util import to_relative

from .result_status import ResultStatus

logger = getLogger(__name__)


class VerificationResult(BaseModel):
    status: ResultStatus
    elapsed: float
    """
    Elapsed seconds
    """

    last_execution_time: datetime.datetime = Field(
        default_factory=datetime.datetime.now
    )

    @validator("status", pre=True)
    def verification_list(cls, v: Any) -> Any:
        if isinstance(v, str):
            return v.lower()
        return v

    def need_reverifying(self, base_time: datetime.datetime) -> bool:
        if self.status != ResultStatus.SUCCESS:
            return True
        if self.last_execution_time is None:
            return True

        return self.last_execution_time < base_time


class FileResult(BaseModel):
    verifications: list[VerificationResult] = Field(default_factory=list)
    newest: bool = True

    def need_verification(self, base_time: datetime.datetime) -> bool:
        if len(self.verifications) == 0:
            return True
        return any(r.need_reverifying(base_time) for r in self.verifications)

    def is_success(self, allow_skip: bool) -> bool:
        if allow_skip:
            return all(r.status != ResultStatus.FAILURE for r in self.verifications)
        else:
            return all(r.status == ResultStatus.SUCCESS for r in self.verifications)


class VerifyCommandResult(BaseModel):
    total_seconds: float
    files: dict[pathlib.Path, FileResult] = Field(default_factory=dict)

    @classmethod
    def parse_file_relative(
        cls, path: Union[str, pathlib.Path], **kwargs: Any
    ) -> "VerifyCommandResult":
        impl = cls.parse_file(path, **kwargs)
        new_files: dict[pathlib.Path, FileResult] = {}
        for p, f in impl.files.items():
            p = to_relative(p)
            if not p:
                continue
            new_files[p] = f

        impl.files = new_files
        return impl

    def json(self, **kwargs: Any) -> str:
        class WithStrDict(BaseModel):
            total_seconds: float
            files: dict[str, FileResult]

            class Config:
                json_encoders = {pathlib.Path: lambda v: v.as_posix()}  # type: ignore

        return WithStrDict(
            total_seconds=self.total_seconds,
            files={k.as_posix(): v for k, v in self.files.items()},
        ).json(**kwargs)

    def merge(self, other: "VerifyCommandResult") -> "VerifyCommandResult":
        d = self.files.copy()
        for k, r in other.files.items():
            cur = d.get(k)
            if r.newest or (cur is None) or (not cur.newest):
                d[k] = r
        return VerifyCommandResult(
            total_seconds=self.total_seconds + other.total_seconds,
            files=d,
        )

    def is_success(self, allow_skip: bool = True) -> bool:
        return all(f.is_success(allow_skip) for f in self.files.values())



from typing import Optional


class VerifierError(Exception):
    def __init__(
        self,
        message: str,
        *,
        inner: Optional[BaseException] = None,
    ) -> None:
        self.message = message
        self.inner = inner



import sys


def ulimit_stack() -> None:
    """ulimit -s unlimited"""
    if sys.platform != "win32":
        import resource

        _, hard = resource.getrlimit(resource.RLIMIT_STACK)
        resource.setrlimit(resource.RLIMIT_STACK, (hard, hard))



import importlib.resources
import pathlib
from functools import cache
from logging import getLogger
from typing import Any, Iterable, Optional, Union

import yaml

from competitive_verifier import git, github, log
from competitive_verifier.models import (
    ProblemVerification,
    SourceCodeStat,
    VerificationInput,
    VerifyCommandResult,
    resolve_dependency,
)
from competitive_verifier.util import read_text_normalized

from .front_matter import merge_front_matter
from .job import build_index_job, build_markdown_job, build_source_job
from .render import load_render_config
from .type import PageRenderJob, SiteRenderConfig

logger = getLogger(__name__)


_RESOURCE_PACKAGE = "competitive_verifier_resources"
_DOC_USAGE_PATH = "doc_usage.txt"
_RESOURCE_STATIC_FILE_PATHS: list[str] = [
    "_layouts/page.html",
    "_layouts/document.html",
    "_layouts/toppage.html",
    "_includes/head-custom.html",
    "_includes/head-custom2.html",
    "_includes/mathjax/mathjax.html",
    "_includes/mathjax/mathjax2.html",
    "_includes/mathjax/mathjax3.html",
    "_includes/theme_fix.html",
    "_includes/code.html",
    "_includes/highlight_additional.html",
    "_includes/highlight/highlight_header.html",
    "_includes/document_header.html",
    "_includes/document_body.html",
    "_includes/document_footer.html",
    "_includes/toppage_header.html",
    "_includes/toppage_body.html",
    "assets/css/code.scss",
    "assets/js/code.js",
    "Gemfile",
]


class ExcludedPaths:
    def __init__(self, excluded_paths: list[Union[str, pathlib.Path]]) -> None:
        self.excluded_paths = [pathlib.Path(p) for p in excluded_paths]

    @cache
    def is_excluded(self, relative_path: pathlib.Path) -> bool:
        for excluded in self.excluded_paths:
            if relative_path == excluded or excluded in relative_path.parents:
                return True
        return False


class DocumentBuilder:
    input: VerificationInput
    result: VerifyCommandResult
    docs_dir: Optional[pathlib.Path]
    destination_dir: pathlib.Path

    def __init__(
        self,
        input: VerificationInput,
        result: VerifyCommandResult,
        docs_dir: Optional[pathlib.Path],
        destination_dir: pathlib.Path,
    ) -> None:
        self.input = input
        self.result = result
        self.docs_dir = docs_dir
        self.destination_dir = destination_dir

    def build(self) -> bool:
        if not _working_directory_is_in_git_root():
            logger.info(
                "Working directory %s is not git root.",
                pathlib.Path.cwd().as_posix(),
            )

        logger.info("Generate documents...")
        result = self.impl()
        logger.info("Generated.")
        logger.info(
            (importlib.resources.files(_RESOURCE_PACKAGE) / _DOC_USAGE_PATH)
            .read_text(encoding="utf-8")
            .replace("{{{{{markdown_dir_path}}}}}", self.destination_dir.as_posix())
            .replace(
                "{{{{{repository}}}}}",
                github.env.get_repository()
                or "competitive-verifier/competitive-verifier",
            )
        )

        return result

    def impl(self) -> bool:
        config = load_render_config(
            docs_dir=self.docs_dir,
            destination_dir=self.destination_dir,
        )
        logger.debug("lender config=%s", config)

        excluded_paths = ExcludedPaths(config.config_yml.get("exclude", []))
        included_files = set(
            f for f in git.ls_files() if not excluded_paths.is_excluded(f)
        )
        logger.debug(
            "included_files=%s",
            " ".join(p.as_posix() for p in included_files),
        )

        render_jobs = self.enumerate_rendering_jobs(config.index_md, included_files)

        logger.info("render %s files...", len(render_jobs))
        with log.group("Resolve dependency"):
            stats = resolve_dependency(
                input=self.input,
                result=self.result,
                included_files=included_files,
            )
        rendered_pages = self.render_pages(
            stats=stats,
            render_jobs=render_jobs,
            site_render_config=config,
            excluded_paths=excluded_paths,
        )

        # make install
        logger.info("writing %s files...", len(rendered_pages))
        for path, content in rendered_pages.items():
            path.parent.mkdir(parents=True, exist_ok=True)
            logger.debug("writing to %s", path.as_posix())
            path.write_bytes(content)

        logger.info("list static files...")
        static_files = load_static_files(config=config)

        logger.info("writing %s static files...", len(static_files))
        for path, content in static_files.items():
            path.parent.mkdir(parents=True, exist_ok=True)
            logger.debug("writing to %s", path.as_posix())
            path.write_bytes(content)
        return True

    def render_pages(
        self,
        *,
        stats: dict[pathlib.Path, SourceCodeStat],
        render_jobs: list[PageRenderJob],
        site_render_config: SiteRenderConfig,
        excluded_paths: ExcludedPaths,
    ) -> dict[pathlib.Path, bytes]:
        """
        :returns: mapping from absolute paths to file contents
        """

        page_title_dict = _build_page_title_dict(render_jobs=render_jobs)

        rendered_pages: dict[pathlib.Path, bytes] = {}
        for job in render_jobs:
            documentation_of = job.front_matter.documentation_of

            front_matter = job.front_matter.copy(deep=True)
            if front_matter.layout == "toppage":
                front_matter.data = _render_source_code_stats_for_top_page(
                    stats_iter=stats.values(),
                    page_title_dict=page_title_dict,
                )

            elif documentation_of is not None:
                documentation_of_path = pathlib.Path(documentation_of)
                if not documentation_of_path.exists():
                    logger.warning(
                        "Skip %s because %s doesn't exist.",
                        job.path,
                        documentation_of,
                    )
                    continue
                if excluded_paths.is_excluded(documentation_of_path):
                    logger.info(
                        "Skip %s because %s is excluded.",
                        job.path,
                        documentation_of,
                    )
                    continue
                if not front_matter.layout:
                    front_matter.layout = "document"
                if not front_matter.data:
                    front_matter.data = _render_source_code_stat_for_page(
                        job,
                        documentation_of_path,
                        stats=stats,
                        page_title_dict=page_title_dict,
                    )

            path = site_render_config.destination_dir / job.path
            content = merge_front_matter(front_matter, job.content)
            rendered_pages[path] = content

        return rendered_pages

    def enumerate_rendering_jobs(
        self,
        index_md: pathlib.Path,
        included_files: set[pathlib.Path],
    ) -> list[PageRenderJob]:
        markdown_paths = set(
            p
            for p in git.ls_files()
            if p.suffix == ".md" and not p.name.startswith(".")
        )
        markdown_paths &= included_files
        logger.debug(
            "markdown_paths=%s",
            " ".join(p.as_posix() for p in markdown_paths),
        )

        logger.info("list rendering jobs...")

        source_jobs = {
            job.path: job
            for job in (build_source_job(p, f) for p, f in self.input.files.items())
            if job is not None
        }
        logger.debug("source_jobs=%s", source_jobs.values())

        source_paths = set(self.input.files.keys())
        markdown_jobs = {
            job.path: job
            for job in map(
                lambda p: build_markdown_job(p, source_paths=source_paths),
                markdown_paths,
            )
            if job is not None
        }
        logger.debug("markdown_jobs=%s", markdown_jobs.values())

        index_job = build_index_job(index_md)
        logger.debug("index_job=%s", index_job)

        result: dict[pathlib.Path, PageRenderJob] = {index_job.path: index_job}
        for path, job in source_jobs.items():
            prev = result.get(path)
            if prev:
                result[path] = job.merge(prev)
            else:
                result[path] = job
        for path, job in markdown_jobs.items():
            prev = result.get(path)
            if prev:
                result[path] = job.merge(prev)
            else:
                result[path] = job
        return list(result.values())


def _working_directory_is_in_git_root() -> bool:
    working_directory = pathlib.Path.cwd()
    if github.env.is_in_github_actions():
        workspace = github.env.get_workspace_path()
        assert workspace is not None
        return working_directory.resolve() == workspace.resolve()
    else:
        return working_directory.resolve() == git.get_root_directory().resolve()


def _build_page_title_dict(
    *, render_jobs: list[PageRenderJob]
) -> dict[pathlib.Path, str]:
    page_title_dict: dict[pathlib.Path, str] = {}
    for job in render_jobs:
        assert job.path.suffix == ".md"
        title = job.front_matter.title or job.path.stem
        page_title_dict[job.path] = title
        page_title_dict[job.path.with_suffix("")] = title
    return page_title_dict


def _render_source_code_stats_for_top_page(
    *,
    stats_iter: Iterable[SourceCodeStat],
    page_title_dict: dict[pathlib.Path, str],
) -> dict[str, Any]:
    library_categories: dict[str, list[dict[str, str]]] = {}
    verification_categories: dict[str, list[dict[str, str]]] = {}
    for stat in stats_iter:
        if stat.is_verification:
            categories = verification_categories
        else:
            categories = library_categories
        category = stat.path.parent.as_posix()
        if category not in categories:
            categories[category] = []
        categories[category].append(
            {
                "path": stat.path.as_posix(),
                "title": page_title_dict[stat.path],
                "icon": stat.verification_status.icon,
            }
        )

    def _build_categories_list(
        categories: dict[str, list[dict[str, str]]]
    ) -> list[dict[str, Any]]:
        return sorted(
            (
                {
                    "name": category,
                    "pages": sorted(pages, key=lambda p: p.get("path", "")),
                }
                for category, pages in categories.items()
            ),
            key=lambda d: d.get("name", ""),
        )

    return {
        "libraryCategories": _build_categories_list(library_categories),
        "verificationCategories": _build_categories_list(verification_categories),
    }


def _render_source_code_stat(stat: SourceCodeStat) -> dict[str, Any]:
    code = read_text_normalized(stat.path)

    attributes = stat.file_input.document_attributes.copy()
    problem = next(
        (
            v.problem
            for v in stat.file_input.verification
            if isinstance(v, ProblemVerification)
        ),
        None,
    )
    if problem:
        attributes.setdefault("PROBLEM", problem)

    embedded = [{"name": "default", "code": code}]
    for s in stat.file_input.additonal_sources:
        embedded.append({"name": s.name, "code": read_text_normalized(s.path)})
    return {
        "path": stat.path.as_posix(),
        "embedded": embedded,
        "isVerificationFile": stat.is_verification,
        "verificationStatus": stat.verification_status.value,
        "timestamp": str(stat.timestamp),
        "dependsOn": [path.as_posix() for path in stat.depends_on],
        "requiredBy": [path.as_posix() for path in stat.required_by],
        "verifiedWith": [path.as_posix() for path in stat.verified_with],
        "attributes": attributes,
    }


def _render_source_code_stat_for_page(
    job: PageRenderJob,
    path: pathlib.Path,
    *,
    stats: dict[pathlib.Path, SourceCodeStat],
    page_title_dict: dict[pathlib.Path, str],
) -> dict[str, Any]:
    stat = stats[path]
    data = _render_source_code_stat(stat)
    data["_pathExtension"] = path.suffix.lstrip(".")
    data["_verificationStatusIcon"] = stat.verification_status.icon
    data["_isVerificationFailed"] = stat.verification_status.is_failed
    if job.document_path:
        data["_document_path"] = job.document_path.as_posix()

    def ext(path: pathlib.Path) -> dict[str, Any]:
        return {
            "path": path.as_posix(),
            "title": page_title_dict[path],
            "icon": stats[path].verification_status.icon,
        }

    def path_list(paths: Iterable[pathlib.Path]) -> list[dict[str, Any]]:
        return [ext(path) for path in sorted(paths, key=str)]

    data["_extendedDependsOn"] = path_list(stat.depends_on)
    data["_extendedRequiredBy"] = path_list(stat.required_by)
    data["_extendedVerifiedWith"] = path_list(stat.verified_with)

    return data


def load_static_files(*, config: SiteRenderConfig) -> dict[pathlib.Path, bytes]:
    files: dict[pathlib.Path, bytes] = {}

    # write merged config.yml
    files[config.destination_dir / "_config.yml"] = yaml.safe_dump(
        config.config_yml
    ).encode()

    # load files in resource/*
    for path in _RESOURCE_STATIC_FILE_PATHS:
        files[config.destination_dir / path] = (
            importlib.resources.files(_RESOURCE_PACKAGE) / "jekyll" / path
        ).read_bytes()

    # overwrite with docs/static
    for src in config.static_dir.glob("**/*"):
        if src.is_file():
            dst = config.destination_dir / src.relative_to(config.static_dir)
            files[dst] = src.read_bytes()
    return files



# Python Version: 3.x
import pathlib
from logging import getLogger
from typing import Optional, Sequence

import competitive_verifier_oj_clone.utils as utils
from competitive_verifier_oj_clone.languages.models import Language, LanguageEnvironment
from competitive_verifier_oj_clone.languages.special_comments import (
    list_special_comments,
)

from .. import subprocess2 as subprocess

logger = getLogger(__name__)


class UserDefinedLanguageEnvironment(LanguageEnvironment):
    config: dict[str, str]

    def __init__(self, *, config: dict[str, str]):
        self.config = config

    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> Optional[str]:
        compile = self.config.get("compile")
        if compile:
            return compile.format(
                path=str(path), basedir=str(basedir), tempdir=str(tempdir)
            )
        return None

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        assert "execute" in self.config
        return self.config["execute"].format(
            path=str(path), basedir=str(basedir), tempdir=str(tempdir)
        )


class UserDefinedLanguage(Language):
    extension: str
    config: dict[str, str]

    def __init__(self, *, extension: str, config: dict[str, str]):
        self.extension = extension
        self.config = config

    def list_attributes(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> dict[str, str]:
        if "list_attributes" not in self.config:
            return list_special_comments(path)
        logger.warning(
            '"languages.*.list_attributes" field in .verify-helper/config.toml is now obsoleted'
        )

        command = self.config["list_attributes"].format(
            path=str(path), basedir=str(basedir)
        )
        text = subprocess.run(command, text=False).stdout
        attributes: dict[str, str] = {}
        for line in text.splitlines():
            key, _, value = line.decode().partition(" ")
            attributes[key] = value
        return attributes

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        if "list_dependencies" not in self.config:
            logger.warning(
                "The functionality to list dependencies of .%s file is not implemented yet.",
                self.extension,
            )
            return list(
                utils.glob_with_predicate(
                    lambda path: path.suffix == "." + self.extension
                )
            )

        command = self.config["list_dependencies"].format(
            path=path.as_posix(), basedir=basedir.as_posix()
        )
        text = subprocess.run(command, text=False).stdout
        dependencies = [path]
        for line in text.splitlines():
            dependencies.append(pathlib.Path(line.decode()))
        return dependencies

    def bundle(self, path: pathlib.Path, *, basedir: pathlib.Path) -> Optional[bytes]:
        if "bundle" not in self.config:
            return None
        command = self.config["bundle"].format(path=str(path), basedir=str(basedir))
        logger.info("$ %s", command)
        return subprocess.run(command, text=False).stdout

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> Sequence[LanguageEnvironment]:
        return [UserDefinedLanguageEnvironment(config=self.config)]



import pathlib
from logging import getLogger
from typing import Any, Optional, Sequence

import competitive_verifier_oj_clone.shlex2 as shlex
from competitive_verifier_oj_clone.config import get_config
from competitive_verifier_oj_clone.languages.models import LanguageEnvironment
from competitive_verifier_oj_clone.languages.user_defined import UserDefinedLanguage

logger = getLogger(__name__)


class JavaLanguageEnvironment(LanguageEnvironment):
    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        return shlex.join(["javac", str(basedir / path)])

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        relative_path = (basedir / path).relative_to(basedir)
        class_path = ".".join([*relative_path.parent.parts, relative_path.stem])
        return shlex.join(["java", class_path])


class JavaLanguage(UserDefinedLanguage):
    config: dict[str, Any]

    def __init__(self, *, config: Optional[dict[str, Any]] = None):
        if config is None:
            config = get_config()["languages"].get("java", {})
        assert config is not None
        if "compile" in config:
            raise RuntimeError('You cannot overwrite "compile" for Java language')
        if "execute" in config:
            raise RuntimeError('You cannot overwrite "execute" for Java language')
        super().__init__(extension="java", config=config)

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> Sequence[LanguageEnvironment]:
        return [JavaLanguageEnvironment()]



import pathlib
from logging import getLogger
from typing import Optional

from competitive_verifier.models import VerificationFile

from . import front_matter
from .type import FrontMatter, PageRenderJob

logger = getLogger(__name__)


def get_markdown_paths(*, basedir: pathlib.Path) -> list[pathlib.Path]:
    return [p for p in basedir.glob("**/*.md") if not p.name.startswith(".")]


def resolve_documentation_of(
    documentation_of: str,
    *,
    basepath: pathlib.Path,
    check_exists: bool = True,
) -> Optional[pathlib.Path]:
    if documentation_of.startswith("."):
        # a relative path
        path = basepath.parent / pathlib.Path(documentation_of)
        if path.exists():
            return path
    elif documentation_of.startswith("//"):
        # from the document root
        path = pathlib.Path(documentation_of[2:])
        if path.exists():
            return path

    path = pathlib.Path(documentation_of)
    if path.exists():
        return path

    path = basepath.parent / pathlib.Path(documentation_of)
    if path.exists() or not check_exists:
        return path

    return None


def build_markdown_job(
    path: pathlib.Path,
    *,
    source_paths: set[pathlib.Path],
) -> Optional[PageRenderJob]:
    content = path.read_bytes()

    fm, content = front_matter.split_front_matter(content)
    # move the location if documentation_of field exists
    documentation_of = fm.documentation_of
    if documentation_of is None:
        return PageRenderJob(
            path=path,
            document_path=None,
            front_matter=fm,
            content=content,
        )

    documentation_of_path = resolve_documentation_of(documentation_of, basepath=path)
    if documentation_of_path is None:
        logger.warning(
            "the `documentation_of` path of %s is not found: %s",
            path.as_posix(),
            documentation_of,
        )
        return None
    documentation_of_relative_path = documentation_of_path.resolve().relative_to(
        pathlib.Path.cwd()
    )
    if documentation_of_relative_path not in source_paths:
        logger.warning(
            "the `documentation_of` path of %s is not target: %s",
            path.as_posix(),
            documentation_of,
        )
        return None
    fm.documentation_of = documentation_of_relative_path.as_posix()
    mdpath = documentation_of_relative_path.with_suffix(
        documentation_of_relative_path.suffix + ".md"
    )

    return PageRenderJob(
        path=mdpath,
        document_path=path,
        front_matter=fm,
        content=content,
    )


def build_source_job(
    path: pathlib.Path, file: VerificationFile
) -> Optional[PageRenderJob]:
    mdpath = path.with_suffix(path.suffix + ".md")

    # add redirects from old URLs
    old_directory = "/verify/" if file.is_verification() else "/library/"
    redirect_from = [
        old_directory + path.as_posix(),
        old_directory + path.with_suffix(".html").as_posix(),
    ]

    # add title specified as a attributes
    title = file.document_attributes.get("TITLE")
    if not title:
        title = file.document_attributes.get("document_title", path.as_posix())
    front_matter = FrontMatter(
        documentation_of=path.as_posix(),
        redirect_from=redirect_from,
        title=title,
    )
    # treat @docs path/to.md directives
    content = b""
    # if '_deprecated_at_docs' in stat.attributes:
    #     at_docs_path = pathlib.Path(stat.attributes['_deprecated_at_docs'])
    #     try:
    #         with open(at_docs_path, 'rb') as fh:
    #             content = fh.read()
    #     except FileNotFoundError as e:
    #         logger.exception('failed to read markdown file specified by @docs in %s: %s', str(stat.path), e)

    return PageRenderJob(
        path=mdpath,
        document_path=None,
        front_matter=front_matter,
        content=content,
    )


def build_index_job(index_md_path: pathlib.Path) -> PageRenderJob:
    content = b""
    if index_md_path.exists():
        content = index_md_path.read_bytes()
    return PageRenderJob(
        path=pathlib.Path("index.md"),
        document_path=None,
        front_matter=FrontMatter(layout="toppage"),
        content=content,
    )



import argparse
import pathlib
import sys
from collections import Counter
from functools import reduce
from logging import getLogger
from typing import Iterable, Optional

from competitive_verifier.arg import add_result_json_argument
from competitive_verifier.log import configure_stderr_logging
from competitive_verifier.models import ResultStatus, VerifyCommandResult

logger = getLogger(__name__)


def merge(results: Iterable[VerifyCommandResult]) -> VerifyCommandResult:
    return reduce(lambda a, b: a.merge(b), results)


def run_impl(*result_json: pathlib.Path) -> bool:
    configure_stderr_logging()

    result = merge(map(VerifyCommandResult.parse_file_relative, result_json))

    counter = Counter(
        r.status for fr in result.files.values() for r in fr.verifications
    )

    for st in ResultStatus:
        print(f"{st.value}: {counter.get(st, 0)}")

    if counter[ResultStatus.FAILURE] > 0:
        logger.error("Failure test count: %d", counter[ResultStatus.FAILURE])
        return False
    return True


def run(args: argparse.Namespace) -> bool:
    return run_impl(*args.result_json)


def argument(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    add_result_json_argument(parser)
    return parser


def main(args: Optional[list[str]] = None) -> None:
    try:
        parsed = argument(argparse.ArgumentParser()).parse_args(args)
        if not run(parsed):
            sys.exit(1)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(2)


if __name__ == "__main__":
    main()



import argparse
import os
import pathlib
from typing import Any
from unittest import mock

import pytest

import competitive_verifier.oj as oj


@pytest.fixture
def clear_env():
    os.environ.clear()


def test_oj_download(clear_env: Any):
    with mock.patch(
        "competitive_verifier.oj.get_cache_directory",
        return_value=pathlib.Path("/bar/baz/online-judge-tools"),
    ), mock.patch.object(pathlib.Path, "mkdir") as mkdir, mock.patch(
        "onlinejudge_command.subcommand.download.run"
    ) as patch:
        oj.download("http://example.com")
        mkdir.assert_called_once()
        patch.assert_called_once()
        args = patch.call_args[0][0]

        assert isinstance(args, argparse.Namespace)
        assert args.subcommand == "download"
        assert args.cookie == pathlib.Path("/bar/baz/online-judge-tools") / "cookie.txt"
        assert args.url == "http://example.com"
        assert args.directory == pathlib.Path(
            ".competitive-verifier/cache/problems/a9b9f04336ce0181a08e774e01113b31/test"
        )
        assert args.yukicoder_token is None


def test_oj_download_yukicoder():
    with mock.patch(
        "competitive_verifier.oj.get_cache_directory",
        return_value=pathlib.Path("/bar/baz/online-judge-tools"),
    ), mock.patch.object(pathlib.Path, "mkdir") as mkdir, mock.patch(
        "onlinejudge_command.subcommand.download.run"
    ) as patch:
        try:
            os.environ["YUKICODER_TOKEN"] = "YKTK"
            oj.download("http://example.com")
            mkdir.assert_called_once()
            patch.assert_called_once()
            args = patch.call_args[0][0]

            assert isinstance(args, argparse.Namespace)
            assert args.subcommand == "download"
            assert (
                args.cookie
                == pathlib.Path("/bar/baz/online-judge-tools") / "cookie.txt"
            )
            assert args.url == "http://example.com"
            assert args.directory == pathlib.Path(
                ".competitive-verifier/cache/problems/a9b9f04336ce0181a08e774e01113b31/test"
            )
            assert args.yukicoder_token == "YKTK"
        finally:
            del os.environ["YUKICODER_TOKEN"]


def test_oj_test():
    with mock.patch(
        "competitive_verifier.oj.get_cache_directory",
        return_value=pathlib.Path("/bar/baz/online-judge-tools"),
    ), mock.patch("onlinejudge_command.subcommand.test.run") as patch:
        oj.test(url="http://example.com", command="ls .", tle=2, error=None)

        patch.assert_called_once()
        args = patch.call_args[0][0]

        assert isinstance(args, argparse.Namespace)
        assert args.subcommand == "test"
        assert args.print_input is True
        assert args.cookie == pathlib.Path("/bar/baz/online-judge-tools") / "cookie.txt"
        assert args.tle == 2.0
        assert args.error is None
        assert args.command == "ls ."
        assert args.directory == pathlib.Path(
            ".competitive-verifier/cache/problems/a9b9f04336ce0181a08e774e01113b31/test"
        )



# Python Version: 3.x
import glob
import pathlib
from os import PathLike
from typing import Any, Callable, Iterator

from charset_normalizer import from_path


def glob_with_predicate(pred: Callable[[pathlib.Path], bool]) -> Iterator[pathlib.Path]:
    """glob_with_basename iterates files whose basenames satisfy the predicate.

    This function ignores hidden directories and hidden files, whose names start with dot `.` letter.
    """
    return filter(pred, map(pathlib.Path, glob.glob("**", recursive=True)))


def read_text_normalized(path: PathLike[Any]) -> str:
    return str(from_path(path).best())



import textwrap

import pytest

from competitive_verifier.migrate.migration import problem_subn

test_problem_subn_params = {
    "NO PROBLEM": (
        textwrap.dedent(
            """
        #define PROGRAM 2
        void main() { }"""
        ),
        (
            textwrap.dedent(
                """
        #define PROGRAM 2
        void main() { }"""
            ),
            0,
        ),
    ),
    "PROBLEM": (
        textwrap.dedent(
            """
        #define PROBLEM http://example.com
        void main() { }"""
        ),
        (
            textwrap.dedent(
                """
        // competitive-verifier: PROBLEM http://example.com
        void main() { }"""
            ),
            1,
        ),
    ),
    "BACKSLASH BREAK": (
        textwrap.dedent(
            r"""
        #define PROBLEM \
           http://example.com
        void main() { }"""
        ),
        (
            textwrap.dedent(
                """
        // competitive-verifier: PROBLEM http://example.com
        void main() { }"""
            ),
            1,
        ),
    ),
    "QUOTE PROBLEM": (
        textwrap.dedent(
            """
        #define PROBLEM "http://example.com"
        void main() { }"""
        ),
        (
            textwrap.dedent(
                """
        // competitive-verifier: PROBLEM http://example.com
        void main() { }"""
            ),
            1,
        ),
    ),
    "QUOTE BACKSLASH BREAK": (
        textwrap.dedent(
            r"""
        #define PROBLEM \
           "http://example.com"
        void main() { }"""
        ),
        (
            textwrap.dedent(
                """
        // competitive-verifier: PROBLEM http://example.com
        void main() { }"""
            ),
            1,
        ),
    ),
    "NO BACKSLASH PROBLEM": (
        textwrap.dedent(
            """
        #define PROBLEM
        // http://example.com
        void main() { }"""
        ),
        (
            textwrap.dedent(
                """
        #define PROBLEM
        // http://example.com
        void main() { }"""
            ),
            0,
        ),
    ),
}


@pytest.mark.parametrize(
    "content, expected",
    test_problem_subn_params.values(),
    ids=test_problem_subn_params.keys(),
)
def test_problem_subn(content: str, expected: tuple[str, int]):
    assert problem_subn(content) == expected



import pytest

import competitive_verifier.merge_input.main
from competitive_verifier.models import VerificationInput

test_merge_params: list[tuple[list[VerificationInput], VerificationInput]] = [
    (
        [
            VerificationInput.parse_obj(
                {
                    "files": {
                        "foo/bar.py": {},
                        "foo/baz.py": {
                            "path": "foo/baz.py",
                            "document_attributes": {
                                "title": "foo-baz",
                            },
                            "dependencies": ["foo/bar.py"],
                            "verification": [
                                {
                                    "type": "const",
                                    "status": "success",
                                }
                            ],
                        },
                    },
                }
            ),
        ],
        VerificationInput.parse_obj(
            {
                "files": {
                    "foo/bar.py": {},
                    "foo/baz.py": {
                        "path": "foo/baz.py",
                        "document_attributes": {
                            "title": "foo-baz",
                        },
                        "dependencies": ["foo/bar.py"],
                        "verification": [
                            {
                                "type": "const",
                                "status": "success",
                            }
                        ],
                    },
                },
            }
        ),
    ),
    (
        [
            VerificationInput.parse_obj(
                {
                    "files": {
                        "foo/bar.py": {},
                        "foo/baz.py": {
                            "path": "foo/baz.py",
                            "document_attributes": {
                                "title": "foo-baz",
                            },
                            "dependencies": ["foo/bar.py"],
                            "verification": [
                                {
                                    "type": "const",
                                    "status": "success",
                                }
                            ],
                        },
                    },
                }
            ),
            VerificationInput.parse_obj(
                {
                    "files": {
                        "hoge/bar.py": {},
                        "hoge/baz.py": {
                            "path": "hoge/baz.py",
                            "document_attributes": {
                                "title": "foo-baz",
                            },
                            "dependencies": ["hoge/bar.py"],
                            "verification": [
                                {
                                    "type": "const",
                                    "status": "success",
                                }
                            ],
                        },
                    },
                }
            ),
        ],
        VerificationInput.parse_obj(
            {
                "files": {
                    "foo/bar.py": {},
                    "foo/baz.py": {
                        "path": "foo/baz.py",
                        "document_attributes": {
                            "title": "foo-baz",
                        },
                        "dependencies": ["foo/bar.py"],
                        "verification": [
                            {
                                "type": "const",
                                "status": "success",
                            }
                        ],
                    },
                    "hoge/bar.py": {},
                    "hoge/baz.py": {
                        "path": "hoge/baz.py",
                        "document_attributes": {
                            "title": "foo-baz",
                        },
                        "dependencies": ["hoge/bar.py"],
                        "verification": [
                            {
                                "type": "const",
                                "status": "success",
                            }
                        ],
                    },
                },
            }
        ),
    ),
]


@pytest.mark.parametrize("inputs, expected", test_merge_params)
def test_merge(inputs: list[VerificationInput], expected: VerificationInput):
    assert competitive_verifier.merge_input.main.merge(inputs) == expected



# The file is inspired by Tyrrrz/GitHubActionsTestLogger
# https://github.com/Tyrrrz/GitHubActionsTestLogger/blob/04fe7796a047dbd0e3cd6a46339b2a50f5125317/GitHubActionsTestLogger/TestSummary.cs

import os
import pathlib
from collections import Counter
from itertools import chain
from typing import TextIO

from competitive_verifier.models import FileResult, ResultStatus, VerifyCommandResult

SUCCESS = ResultStatus.SUCCESS
FAILURE = ResultStatus.FAILURE
SKIPPED = ResultStatus.SKIPPED


def to_human_str(total_seconds: float) -> str:
    hours = int(total_seconds // 3600)
    rm = total_seconds % 3600
    minutes = int(rm // 60)
    rm %= 60
    seconds = rm

    if hours > 0:
        return f"{hours}h {minutes}m"
    elif minutes > 0:
        return f"{minutes}m {int(seconds)}s"
    elif total_seconds >= 10:
        return f"{int(seconds)}s"
    elif total_seconds > 1:
        return f"{total_seconds:.1f}s"
    else:
        return f"{int(total_seconds*1000)}ms"


def write_summary(fp: TextIO, result: VerifyCommandResult):
    def with_icon(icon: str, text: str) -> str:
        return icon + "&nbsp;&nbsp;" + text

    def write_table_line(cells: list[str]) -> None:
        for c in cells:
            fp.write("|")
            fp.write(c)
        fp.write("|\n")

    def write_table_file_result(results: list[tuple[pathlib.Path, FileResult]]) -> None:
        for p, fr in results:
            counter = Counter(r.status for r in fr.verifications)
            if counter.get(FAILURE):
                emoji_status = "❌"
            elif counter.get(SKIPPED):
                emoji_status = "⚠"
            else:
                emoji_status = "✔"
            write_table_line(
                [
                    with_icon(emoji_status, p.as_posix()),
                    str(counter.get(SUCCESS, "-")),
                    str(counter.get(FAILURE, "-")),
                    str(counter.get(SKIPPED, "-")),
                    str(sum(counter.values())),
                    to_human_str(sum(r.elapsed for r in fr.verifications)),
                ]
            )

    file_results: dict[bool, list[tuple[pathlib.Path, FileResult]]] = {
        False: [],
        True: [],
    }
    for p, fr in result.files.items():
        file_results[fr.newest].append((p, fr))

    file_results[True].sort(key=lambda t: t[0])
    file_results[False].sort(key=lambda t: t[0])
    counter = Counter(
        r.status
        for r in chain.from_iterable(f[1].verifications for f in file_results[True])
    )

    fp.write("# ")

    if counter.get(FAILURE):
        emoji_status = "❌"
    elif counter.get(SKIPPED):
        emoji_status = "⚠"
    else:
        emoji_status = "✔"

    fp.write(emoji_status)
    fp.write(" ")
    fp.write(os.getenv("COMPETITIVE_VERIFY_SUMMARY_TITLE", "Verification result"))
    fp.write("\n\n")

    fp.write("- ")
    fp.write(with_icon("✔", "All test case results are `success`"))
    fp.write("\n")
    fp.write("- ")
    fp.write(with_icon("❌", "Test case results containts `failure`"))
    fp.write("\n")
    fp.write("- ")
    fp.write(with_icon("⚠", "Test case results containts `skipped`"))
    fp.write("\n\n\n")

    header = [
        with_icon("📝", "File"),
        with_icon("✓", "Passed"),
        with_icon("✘", "Failed"),
        with_icon("↷", "Skipped"),
        with_icon("∑", "Total"),
        with_icon("⧗", "Elapsed"),
    ]
    alignment = [":---"] + [":---:"] * (len(header) - 1)

    if file_results[True]:

        fp.write("## Results\n")
        write_table_line(header)
        write_table_line(alignment)
        write_table_line(
            [
                "_**Sum**_",
                str(counter.get(SUCCESS, "-")),
                str(counter.get(FAILURE, "-")),
                str(counter.get(SKIPPED, "-")),
                str(sum(counter.values())),
                to_human_str(result.total_seconds),
            ]
        )
        write_table_line([""] * len(header))

        write_table_file_result(file_results[True])

    if file_results[False]:
        fp.write("## Past results\n")
        write_table_line(header)
        write_table_line(alignment)
        write_table_file_result(file_results[False])



import abc
import enum
import functools
import itertools
import json
import pathlib
import shutil
from collections import defaultdict
from enum import Enum
from logging import getLogger
from typing import Any, Optional, Sequence

import competitive_verifier_oj_clone.shlex2 as shlex
from competitive_verifier_oj_clone.config import get_config
from competitive_verifier_oj_clone.languages.models import Language, LanguageEnvironment
from competitive_verifier_oj_clone.utils import read_text_normalized

from .. import subprocess2 as subprocess

logger = getLogger(__name__)
_metadata_by_manifest_path: dict[pathlib.Path, dict[str, Any]] = {}
_cargo_checked_workspaces: set[pathlib.Path] = set()
_related_source_files_by_workspace: dict[
    pathlib.Path, dict[pathlib.Path, frozenset[pathlib.Path]]
] = {}


class _ListDependenciesBackend:
    @abc.abstractmethod
    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        raise NotImplementedError


class _NoBackend(_ListDependenciesBackend):
    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        return _list_dependencies_by_crate(
            path, basedir=basedir, cargo_udeps_toolchain=None
        )


class _CargoUdeps(_ListDependenciesBackend):
    toolchain: str = "nightly"

    def __init__(self, *, toolchain: Optional[str]):
        if toolchain is not None:
            self.toolchain = toolchain

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        return _list_dependencies_by_crate(
            path, basedir=basedir, cargo_udeps_toolchain=self.toolchain
        )


@functools.lru_cache(maxsize=None)
def _list_dependencies_by_crate(
    path: pathlib.Path, *, basedir: pathlib.Path, cargo_udeps_toolchain: Optional[str]
) -> list[pathlib.Path]:
    """The `list_dependencies` implementation for `_NoBackend` and `CargoUdeps`.

    :param path: A parameter in `Language.list_dependencies`.
    :param basedir: A parameter in `Language.list_dependencies`.
    :param cargo_udeps_toolchain: A Rust toolchain name for cargo-udeps. If it is `None`, we don't run cargo-udeps.
    :returns: Paths to the `.rs` files for `Language.list_dependencies`.
    """
    path = basedir / path

    # We regard that a generated file does not depend on any files.
    for parent in path.parents:
        if (parent.parent / "Cargo.toml").exists() and parent.parts[-1] == "target":
            logger.warning("This is a generated file!: %s", path)
            return [path]

    metadata = _cargo_metadata(cwd=path.parent)

    # First, collects source files in the same crate.
    common_result = set(
        _source_files_in_same_targets(path, _related_source_files(basedir, metadata))
    )

    main_package_and_target = _find_target(metadata, path)
    if not main_package_and_target:
        return sorted(common_result)
    main_package, main_target = main_package_and_target

    packages_by_id = {p["id"]: p for p in metadata["packages"]}

    class DependencyNamespace(Enum):
        NORMAL_DEVELOPMENT = enum.auto()
        BUILD = enum.auto()

        @classmethod
        def from_dep_kind(cls, kind: str):
            if kind == "build":
                return cls.BUILD
            return cls.NORMAL_DEVELOPMENT

    # Collect the `(|dev-|build-)dependencies` into a <is a `build-dependency`> → (<"extern crate name"> → <package>) dictionary.
    dependencies: defaultdict[
        DependencyNamespace, dict[str, dict[str, Any]]
    ] = defaultdict(dict)
    for dep in next(
        n["deps"] for n in metadata["resolve"]["nodes"] if n["id"] == main_package["id"]
    ):
        if _need_dev_deps(main_target) or any(
            k["kind"] is None for k in dep["dep_kinds"]
        ):
            dependencies[DependencyNamespace.NORMAL_DEVELOPMENT][
                dep["name"]
            ] = packages_by_id[dep["pkg"]]
        if any(k["kind"] == "build" for k in dep["dep_kinds"]):
            dependencies[DependencyNamespace.BUILD][dep["name"]] = packages_by_id[
                dep["pkg"]
            ]

    # If `cargo_udeps_toolchain` is present, collects packages that are "unused" by `target`.
    unused_packages: defaultdict[DependencyNamespace, set[Any]] = defaultdict(set)
    if cargo_udeps_toolchain is not None:
        explicit_names_in_toml = {
            (DependencyNamespace.from_dep_kind(d["kind"]), d["rename"])
            for d in main_package["dependencies"]
            if d["rename"]
        }
        if not shutil.which("cargo-udeps"):
            raise RuntimeError("`cargo-udeps` not in $PATH")
        args: list[str] = [
            "rustup",
            "run",
            cargo_udeps_toolchain,
            "cargo",
            "udeps",
            "--output",
            "json",
            "--manifest-path",
            main_package["manifest_path"],
            *_target_option(main_target),
        ]
        unused_deps = json.loads(
            subprocess.run(
                args,
                cwd=metadata["workspace_root"],
                check=False,
            ).stdout.decode()
        )["unused_deps"].values()
        unused_dep = next(
            (
                u
                for u in unused_deps
                if u["manifest_path"] == main_package["manifest_path"]
            ),
            None,
        )
        if unused_dep:
            names_in_toml: list[tuple[DependencyNamespace, Any]] = [
                (DependencyNamespace.NORMAL_DEVELOPMENT, name_in_toml)
                for name_in_toml in [*unused_dep["normal"], *unused_dep["development"]]
            ]
            names_in_toml.extend(
                (DependencyNamespace.BUILD, name_in_toml)
                for name_in_toml in unused_dep["build"]
            )
            for dependency_namespace, name_in_toml in names_in_toml:
                if (dependency_namespace, name_in_toml) in explicit_names_in_toml:
                    # If the `name_in_toml` is explicitly renamed one, it equals to the `extern_crate_name`.
                    unused_package: Any = dependencies[dependency_namespace][
                        name_in_toml
                    ]["id"]
                else:
                    # Otherwise, it equals to the `package.name`.
                    unused_package = next(
                        p["id"]
                        for p in dependencies[dependency_namespace].values()
                        if p["name"] == name_in_toml
                    )
                unused_packages[dependency_namespace].add(unused_package)

    # Finally, adds source files related to the depended crates except:
    #
    # - those detected by cargo-udeps
    # - those come from Crates.io or Git repositories (e.g. `proconio`, other people's libraries including `ac-library-rs`)

    # `main_package` should always be included.
    # Note that cargo-udeps does not detect it if it is unused.
    # https://github.com/est31/cargo-udeps/pull/35
    depended_packages = [main_package]
    for dependency_namespace, values in dependencies.items():
        for depended_package in values.values():
            if (
                depended_package["id"] not in unused_packages[dependency_namespace]
                and not depended_package["source"]
            ):
                depended_packages.append(depended_package)

    ret = common_result

    for depended_package in depended_packages:
        depended_targets = [
            t
            for t in depended_package["targets"]
            if t != main_target and (_is_build(t) or _is_lib_or_proc_macro(t))
        ]
        assert len(depended_targets) <= 2
        for depended_target in depended_targets:
            related_source_files = _related_source_files(
                basedir,
                _cargo_metadata_by_manifest_path(
                    pathlib.Path(depended_package["manifest_path"])
                ),
            )
            ret |= _source_files_in_same_targets(
                pathlib.Path(depended_target["src_path"]).resolve(strict=True),
                related_source_files,
            )
    return sorted(ret)


def _related_source_files(
    basedir: pathlib.Path, metadata: dict[str, Any]
) -> dict[pathlib.Path, frozenset[pathlib.Path]]:
    """Collects all of the `.rs` files recognized by a workspace.

    :param basedir: A parameter from `Language.list_dependencies`.
    :param metadata: Output of `cargo metadata`
    :returns: A (main source file) → (other related files) map
    """
    if pathlib.Path(metadata["workspace_root"]) in _related_source_files_by_workspace:
        return _related_source_files_by_workspace[
            pathlib.Path(metadata["workspace_root"])
        ]

    # Runs `cargo check` to generate `$target_directory/debug/deps/*.d`.
    if pathlib.Path(metadata["workspace_root"]) not in _cargo_checked_workspaces:
        subprocess.run(
            [
                "cargo",
                "check",
                "--manifest-path",
                str(pathlib.Path(metadata["workspace_root"], "Cargo.toml")),
                "--workspace",
                "--all-targets",
            ],
            cwd=metadata["workspace_root"],
            check=True,
        )
        _cargo_checked_workspaces.add(pathlib.Path(metadata["workspace_root"]))

    ret: dict[pathlib.Path, frozenset[pathlib.Path]] = dict()

    targets_in_workspace = itertools.chain.from_iterable(
        p["targets"]
        for p in metadata["packages"]
        if p["id"] in metadata["workspace_members"]
    )
    for target in targets_in_workspace:
        # Finds the **latest** "dep-info" file that contains a line in the following format, and parses the line.
        #
        # ```
        # <relative/absolute path to the `.d` file itself>: <relative/absolute path to the root source file> <relative/aboslute paths to the other related files>...
        # ```
        #
        # - https://github.com/rust-lang/cargo/blob/rust-1.49.0/src/cargo/core/compiler/fingerprint.rs#L1979-L1997
        # - https://github.com/rust-lang/cargo/blob/rust-1.49.0/src/cargo/core/compiler/fingerprint.rs#L1824-L1830
        if _is_build(target):
            dep_info_paths = pathlib.Path(
                metadata["target_directory"], "debug", "build"
            ).rglob(f"{_crate_name(target)}-*.d")
        elif _is_example(target):
            dep_info_paths = pathlib.Path(
                metadata["target_directory"], "debug", "examples"
            ).glob(f"{_crate_name(target)}-*.d")
        else:
            dep_info_paths = pathlib.Path(
                metadata["target_directory"], "debug", "deps"
            ).glob(f"{_crate_name(target)}-*.d")
        for dep_info_path in sorted(
            dep_info_paths, key=lambda p: p.stat().st_mtime_ns, reverse=True
        ):
            dep_info = read_text_normalized(dep_info_path)
            for line in dep_info.splitlines():
                ss = line.split(": ")
                if (
                    len(ss) == 2
                    and pathlib.Path(metadata["workspace_root"], ss[0]) == dep_info_path
                ):
                    paths: list[pathlib.Path] = []
                    it = iter(ss[1].split())
                    for s in it:
                        while s.endswith("\\"):
                            s = s.rstrip("\\")
                            s += " "
                            s += next(it)
                        path = pathlib.Path(metadata["workspace_root"], s).resolve(
                            strict=True
                        )
                        # Ignores paths that don't start with the `basedir`. (e.g. `/dev/null`, `/usr/local/share/foo/bar`)
                        try:
                            # `PurePath.is_relative_to` is since Python 3.9.
                            _ = path.relative_to(basedir)
                            paths.append(path)
                        except ValueError:
                            pass
                    if paths[:1] == [
                        pathlib.Path(target["src_path"]).resolve(strict=True)
                    ]:
                        ret[paths[0]] = frozenset(paths[1:])
                        break
            else:
                continue
            break
        else:
            logger.error("no `.d` file for `%s`", target["name"])

    _related_source_files_by_workspace[pathlib.Path(metadata["workspace_root"])] = ret
    return ret


def _source_files_in_same_targets(
    path: pathlib.Path,
    related_source_files: dict[pathlib.Path, frozenset[pathlib.Path]],
) -> frozenset[pathlib.Path]:
    """Returns `.rs` file paths relating to `path`.

    :param path: Path to a `.rs` file
    :param related_source_files: Output of `_related_source_files`
    :returns: Relating `.rs` file paths
    """
    # If `p` is `src_path` of a target, it does not belong to any other target unless it's weirdly symlinked,
    if path in related_source_files:
        return frozenset({path, *related_source_files[path]})

    # Otherwise, it may be used by multiple targets with `#[path = ".."] mod foo;` or something.
    return frozenset(
        itertools.chain.from_iterable(
            {k, *v} for (k, v) in related_source_files.items() if path in v
        )
    ) or frozenset({path})


class RustLanguageEnvironment(LanguageEnvironment):
    def get_compile_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> Optional[str]:
        path = basedir / path
        metadata = _cargo_metadata(cwd=path.parent)
        target = _ensure_target(metadata, path)
        return f"cd {str(path.parent.resolve())} && " + shlex.join(
            ["cargo", "build", "--release", *_target_option(target)]
        )

    def get_execute_command(
        self, path: pathlib.Path, *, basedir: pathlib.Path, tempdir: pathlib.Path
    ) -> str:
        path = basedir / path
        metadata = _cargo_metadata(cwd=path.parent)
        target = _ensure_target(metadata, path)
        return str(
            pathlib.Path(
                metadata["target_directory"],
                "release",
                *([] if _is_bin(target) else ["examples"]),
                target["name"],
            )
        )


class RustLanguage(Language):
    _list_dependencies_backend: _ListDependenciesBackend

    def __init__(self, *, config: Optional[dict[str, Any]] = None):
        if config is None:
            config = get_config()["languages"].get("rust", {})

        assert config is not None
        # Parses `languages.rust.list_dependencies_backend`.
        if "list_dependencies_backend" in config:
            list_dependencies_backend = config["list_dependencies_backend"]

            if not isinstance(list_dependencies_backend, dict):
                raise RuntimeError(
                    "`languages.rust.list_dependencies_backend` must be `dict`"
                )
            if "kind" not in list_dependencies_backend:
                raise RuntimeError(
                    "missing `languages.rust.list_dependencies_backend.kind`"
                )

            list_dependencies_backend_kind: Any = list_dependencies_backend["kind"]

            if not isinstance(list_dependencies_backend_kind, str):
                raise RuntimeError(
                    "`languages.rust.list_dependencies_backend.kind` must be `str`"
                )

            if list_dependencies_backend_kind == "none":
                self._list_dependencies_backend = _NoBackend()
            elif list_dependencies_backend_kind == "cargo-udeps":
                if "toolchain" not in list_dependencies_backend:
                    toolchain = None
                elif isinstance(list_dependencies_backend["toolchain"], str):
                    toolchain = list_dependencies_backend["toolchain"]
                else:
                    raise RuntimeError(
                        "`languages.rust.list_dependencies_backend.toolchain` must be `str`"
                    )
                self._list_dependencies_backend = _CargoUdeps(toolchain=toolchain)
            else:
                raise RuntimeError(
                    "expected 'none' or 'cargo-udeps' for `languages.rust.list_dependencies_backend.kind`"
                )
        else:
            self._list_dependencies_backend = _NoBackend()

    def list_dependencies(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> list[pathlib.Path]:
        return self._list_dependencies_backend.list_dependencies(path, basedir=basedir)

    def list_environments(
        self, path: pathlib.Path, *, basedir: pathlib.Path
    ) -> Sequence[RustLanguageEnvironment]:
        return [RustLanguageEnvironment()]


def _cargo_metadata(cwd: pathlib.Path) -> dict[str, Any]:
    """Returns "metadata" for a Cargo.toml file in `cwd` or its parent directories.

    :raises ValueError: if `cwd` is not absolute or contains `..`
    :returns: Output of `cargo metadata` command
    """
    if not cwd.is_absolute() or ".." in cwd.parts:
        raise ValueError(
            f"the `cwd` parameter must be absolute and must not contain `..`: {cwd}"
        )

    # https://docs.rs/cargo/0.49.0/src/cargo/util/important_paths.rs.html#6-20
    for directory in [cwd, *cwd.parents]:
        manifest_path = directory / "Cargo.toml"
        if manifest_path.exists():
            return _cargo_metadata_by_manifest_path(manifest_path)
    raise RuntimeError(
        f"could not find `Cargo.toml` in `{cwd}` or any parent directory"
    )


def _cargo_metadata_by_manifest_path(manifest_path: pathlib.Path) -> dict[str, Any]:
    """Returns "metadata" for a certain `Cargo.toml`.

    :returns: Output of `cargo metadata` command
    """
    if manifest_path in _metadata_by_manifest_path:
        return _metadata_by_manifest_path[manifest_path]

    metadata = _run_cargo_metadata(manifest_path)
    root_manifest_path = pathlib.Path(metadata["workspace_root"], "Cargo.toml")
    if root_manifest_path != manifest_path:
        metadata = _run_cargo_metadata(root_manifest_path)

    for key in [
        root_manifest_path,
        *(
            pathlib.Path(p["manifest_path"])
            for p in metadata["packages"]
            if p["id"] in metadata["workspace_members"]
        ),
    ]:
        _metadata_by_manifest_path[key] = metadata

    return metadata


def _run_cargo_metadata(manifest_path: pathlib.Path) -> dict[str, Any]:
    """Runs `cargo metadata` for a certain `Cargo.toml`.

    This function is considered to be executed just once for every Cargo.toml in the repository.
    For detailed information about `cargo metadata`, see:

    - <https://doc.rust-lang.org/cargo/commands/cargo-metadata.html#output-format>
    - <https://docs.rs/cargo_metadata>

    :param manifest_path: Path to a `Cargo.toml`
    :returns: Output of `cargo metadata` command
    """
    return json.loads(
        subprocess.run(
            [
                "cargo",
                "metadata",
                "--format-version",
                "1",
                "--manifest-path",
                str(manifest_path),
            ],
            cwd=manifest_path.parent,
            check=True,
        ).stdout.decode()
    )


def _find_target(
    metadata: dict[str, Any],
    src_path: pathlib.Path,
) -> Optional[tuple[dict[str, Any], dict[str, Any]]]:
    for package in metadata["packages"]:
        for target in package["targets"]:
            # A `src_path` may contain `..`
            # The path may not actually exist by being excluded from the package.
            if pathlib.Path(target["src_path"]).resolve() == src_path:
                return package, target
    return None


def _ensure_target(metadata: dict[str, Any], src_path: pathlib.Path) -> dict[str, Any]:
    package_and_target = _find_target(metadata, src_path)
    if not package_and_target:
        raise RuntimeError(f"{src_path} is not a main source file of any target")
    _, target = package_and_target
    return target


def _crate_name(target: dict[str, Any]) -> bool:
    return target["name"].replace("-", "_")


def _is_build(target: dict[str, Any]) -> bool:
    return target["kind"] == ["custom-build"]


def _is_lib_or_proc_macro(target: dict[str, Any]) -> bool:
    return target["kind"] in [["lib"], ["proc-macro"]]


def _is_bin(target: dict[str, Any]) -> bool:
    return target["kind"] == ["bin"]


def _is_example(target: dict[str, Any]) -> bool:
    return target["kind"] == ["example"]


def _need_dev_deps(target: dict[str, Any]) -> bool:
    # Comes from https://docs.rs/cargo/0.49.0/cargo/ops/enum.CompileFilter.html#method.need_dev_deps
    return not (_is_lib_or_proc_macro(target) or _is_bin(target))


def _target_option(target: dict[str, Any]) -> list[str]:
    if target["kind"] == ["bin"]:
        return ["--bin", target["name"]]
    if target["kind"] == ["example"]:
        return ["--example", target["name"]]
    if target["kind"] == ["test"]:
        return ["--test", target["name"]]
    if target["kind"] == ["bench"]:
        return ["--bench", target["name"]]
    return ["--lib"]



import argparse
import logging
import pathlib
import sys
from logging import getLogger
from typing import Optional

import competitive_verifier.merge_result.main as merge_result
from competitive_verifier.arg import (
    add_ignore_error_argument,
    add_result_json_argument,
    add_verify_files_json_argument,
    add_write_summary_argument,
)
from competitive_verifier.config import config_dir
from competitive_verifier.log import configure_logging
from competitive_verifier.models import VerificationInput, VerifyCommandResult

from .builder import DocumentBuilder

logger = getLogger(__name__)


def run_impl(
    input: VerificationInput,
    result: VerifyCommandResult,
    docs_dir: Optional[pathlib.Path],
    destination_dir: pathlib.Path,
    ignore_error: bool,
) -> bool:
    logger.debug("input=%s", input.json())
    logger.debug("result=%s", result.json())
    builder = DocumentBuilder(input, result, docs_dir, destination_dir)
    return builder.build() or ignore_error


def run(args: argparse.Namespace) -> bool:
    logger.debug("arguments=%s", vars(args))
    logger.info("verify_files_json=%s", str(args.verify_files_json))
    logger.info("result_json=%s", [str(p) for p in args.result_json])

    input = VerificationInput.parse_file_relative(args.verify_files_json)
    result = merge_result.run_impl(
        *args.result_json,
        write_summary=args.write_summary,
    )
    return run_impl(input, result, args.docs, args.destination, args.ignore_error)


def argument(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    add_verify_files_json_argument(parser)
    add_result_json_argument(parser)
    add_ignore_error_argument(parser)
    add_write_summary_argument(parser)
    destination = config_dir / "_jekyll"
    docs = config_dir / "docs"
    parser.add_argument(
        "--docs",
        type=pathlib.Path,
        help=f"Document settings directory. default: {docs.as_posix()}",
    )
    parser.add_argument(
        "--destination",
        type=pathlib.Path,
        default=destination,
        help=f"Output directory for markdown document. default: {destination.as_posix()}",
    )
    return parser


def main(args: Optional[list[str]] = None) -> None:
    try:
        configure_logging(logging.INFO)
        parsed = argument(argparse.ArgumentParser()).parse_args(args)
        if not run(parsed):
            sys.exit(1)
    except Exception as e:
        sys.stderr.write(str(e))
        sys.exit(2)



import pathlib
from logging import getLogger
from typing import Optional

from competitive_verifier_oj_clone.config import get_config
from competitive_verifier_oj_clone.languages.models import Language

logger = getLogger(__name__)

_dict: Optional[dict[str, Language]] = None


def _get_dict() -> dict[str, Language]:
    from competitive_verifier_oj_clone.languages.cplusplus import CPlusPlusLanguage
    from competitive_verifier_oj_clone.languages.go import GoLanguage
    from competitive_verifier_oj_clone.languages.haskell import HaskellLanguage
    from competitive_verifier_oj_clone.languages.java import JavaLanguage
    from competitive_verifier_oj_clone.languages.nim import NimLanguage
    from competitive_verifier_oj_clone.languages.python import PythonLanguage
    from competitive_verifier_oj_clone.languages.ruby import RubyLanguage
    from competitive_verifier_oj_clone.languages.rust import RustLanguage
    from competitive_verifier_oj_clone.languages.user_defined import UserDefinedLanguage

    global _dict  # pylint: disable=invalid-name
    if _dict is None:
        _dict = {}
        _dict[".cpp"] = CPlusPlusLanguage()
        _dict[".hpp"] = _dict[".cpp"]
        _dict[".cc"] = _dict[".cpp"]
        _dict[".h"] = _dict[".cpp"]
        _dict[".nim"] = NimLanguage()
        _dict[".py"] = PythonLanguage()
        _dict[".hs"] = HaskellLanguage()
        _dict[".ruby"] = RubyLanguage()
        _dict[".go"] = GoLanguage()
        _dict[".java"] = JavaLanguage()
        _dict[".rs"] = RustLanguage()

        for ext, config in get_config()["languages"].items():
            if "." + ext in _dict:
                if not isinstance(_dict["." + ext], UserDefinedLanguage):
                    for key in (
                        "compile",
                        "execute",
                        "bundle",
                        "list_attributes",
                        "list_dependencies",
                    ):
                        if key in config:
                            raise RuntimeError(
                                "You cannot overwrite existing language: .{}".format(
                                    ext
                                )
                            )
            else:
                _dict["." + ext] = UserDefinedLanguage(extension=ext, config=config)
    return _dict


def get(path: pathlib.Path) -> Optional[Language]:
    return _get_dict().get(path.suffix)



"""
Copy from ac-library-python
https://github.com/not522/ac-library-python/blob/58f324ec020d57191e7b9e4957b0c5feb5ed3aff/atcoder/_scc.py
"""

import sys


class CSR:
    def __init__(self, n: int, edges: list[tuple[int, int]]) -> None:
        self.start = [0] * (n + 1)
        self.elist = [0] * len(edges)

        for e in edges:
            self.start[e[0] + 1] += 1

        for i in range(1, n + 1):
            self.start[i] += self.start[i - 1]

        counter = self.start.copy()
        for e in edges:
            self.elist[counter[e[0]]] = e[1]
            counter[e[0]] += 1


class SccGraph:
    """
    Reference:
    R. Tarjan,
    Depth-First Search and Linear Graph Algorithms
    """

    def __init__(self, n: int) -> None:
        self._n = n
        self._edges: list[tuple[int, int]] = []

    def num_vertices(self) -> int:
        return self._n

    def add_edge(self, from_vertex: int, to_vertex: int) -> None:
        self._edges.append((from_vertex, to_vertex))

    def scc_ids(self) -> tuple[int, list[int]]:
        g = CSR(self._n, self._edges)
        now_ord = 0
        group_num = 0
        visited: list[int] = []
        low = [0] * self._n
        order = [-1] * self._n
        ids = [0] * self._n

        sys.setrecursionlimit(max(self._n + 1000, sys.getrecursionlimit()))

        def dfs(v: int) -> None:
            nonlocal now_ord
            nonlocal group_num
            nonlocal visited
            nonlocal low
            nonlocal order
            nonlocal ids

            low[v] = now_ord
            order[v] = now_ord
            now_ord += 1
            visited.append(v)
            for i in range(g.start[v], g.start[v + 1]):
                to = g.elist[i]
                if order[to] == -1:
                    dfs(to)
                    low[v] = min(low[v], low[to])
                else:
                    low[v] = min(low[v], order[to])

            if low[v] == order[v]:
                while True:
                    u = visited[-1]
                    visited.pop()
                    order[u] = self._n
                    ids[u] = group_num
                    if u == v:
                        break
                group_num += 1

        for i in range(self._n):
            if order[i] == -1:
                dfs(i)

        for i in range(self._n):
            ids[i] = group_num - 1 - ids[i]

        return group_num, ids

    def scc(self) -> list[list[int]]:
        ids = self.scc_ids()
        group_num = ids[0]
        counts = [0] * group_num
        for x in ids[1]:
            counts[x] += 1
        groups: list[list[int]] = [[] for _ in range(group_num)]
        for i in range(self._n):
            groups[ids[1][i]].append(i)

        return groups



# pyright: reportPrivateUsage=none
import datetime
from pathlib import Path
from typing import Any, Optional
from unittest.mock import patch

import pytest

from competitive_verifier.models import (
    ConstVerification,
    FileResult,
    ResultStatus,
    VerificationInput,
    VerificationResult,
    VerifyCommandResult,
)
from competitive_verifier.verify.verifier import BaseVerifier, SplitState

SUCCESS = ResultStatus.SUCCESS
FAILURE = ResultStatus.FAILURE


class NotSkippableConstVerification(ConstVerification):
    @property
    def is_skippable(self) -> bool:
        return False


class MockVerifier(BaseVerifier):
    def __init__(
        self,
        obj: Any = None,
        *,
        prev_result: Optional[VerifyCommandResult],
        verification_time: datetime.datetime,
        split_state: Optional[SplitState],
    ) -> None:
        super().__init__(
            input=VerificationInput.parse_obj(obj),
            verification_time=verification_time,
            prev_result=prev_result,
            split_state=split_state,
            default_tle=10,
            timeout=10,
        )
        self.mock_current_time = datetime.datetime(2006, 1, 2, 15, 4, 5)

    def now(self) -> datetime.datetime:
        dt = self.mock_current_time
        self.mock_current_time = dt + datetime.timedelta(seconds=1)
        return dt

    def get_file_timestamp(self, path: Path) -> datetime.datetime:
        return datetime.datetime(2005, 1, 2, 15, 4, 5)


test_verify_params: list[tuple[MockVerifier, dict[str, Any]]] = [
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/skip.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            ConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            prev_result=None,
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
            split_state=None,
        ),
        {
            "total_seconds": 6.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/skip.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo1.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo2.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo3.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/skip.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            ConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            prev_result=None,
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
            split_state=SplitState(size=2, index=0),
        ),
        {
            "total_seconds": 9.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/foo1.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/skip.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo1.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo2.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo3.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/skip.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            ConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            prev_result=None,
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
            split_state=SplitState(size=2, index=1),
        ),
        {
            "total_seconds": 7.0,
            "files": {
                "test/foo2.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/foo3.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
    (
        MockVerifier(
            {
                "files": {
                    "lib/hoge1.py": {},
                    "test/foo.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo1.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo2.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/foo3.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            NotSkippableConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                    "test/skip.py": {
                        "dependencies": ["lib/hoge1.py"],
                        "verification": [
                            ConstVerification(status=ResultStatus.SUCCESS)
                        ],
                    },
                }
            },
            prev_result=VerifyCommandResult.parse_obj(
                {
                    "total_seconds": 7.0,
                    "files": {
                        "test/foo.py": FileResult(
                            newest=True,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2005, 12, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                        "test/foo1.py": FileResult(
                            newest=True,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2005, 12, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                        "test/foo2.py": FileResult(
                            newest=True,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2005, 12, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                        "test/foo3.py": FileResult(
                            newest=True,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2000, 1, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                        "test/skip.py": FileResult(
                            newest=False,
                            verifications=[
                                VerificationResult(
                                    status=ResultStatus.SUCCESS,
                                    elapsed=1.0,
                                    last_execution_time=datetime.datetime(
                                        2000, 1, 2, 15, 4, 5
                                    ),
                                ),
                            ],
                        ),
                    },
                }
            ),
            verification_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
            split_state=None,
        ),
        {
            "total_seconds": 6.0,
            "files": {
                "test/foo.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(
                                2005, 12, 2, 15, 4, 5
                            ),
                        ),
                    ],
                ),
                "test/foo1.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(
                                2005, 12, 2, 15, 4, 5
                            ),
                        ),
                    ],
                ),
                "test/foo2.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(
                                2005, 12, 2, 15, 4, 5
                            ),
                        ),
                    ],
                ),
                "test/foo3.py": FileResult(
                    newest=True,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
                "test/skip.py": FileResult(
                    newest=False,
                    verifications=[
                        VerificationResult(
                            status=ResultStatus.SUCCESS,
                            elapsed=1.0,
                            last_execution_time=datetime.datetime(2007, 1, 2, 15, 4, 5),
                        ),
                    ],
                ),
            },
        },
    ),
]


@pytest.mark.parametrize("verifier, expected", test_verify_params)
def test_verify(verifier: MockVerifier, expected: Any):
    with patch.object(Path, "exists") as mock_exists:
        mock_exists.return_value = True
        assert verifier.verify() == VerifyCommandResult.parse_obj(expected)



import datetime
import enum
import pathlib

from pydantic import BaseModel

from competitive_verifier import git

from .file import VerificationFile, VerificationInput
from .result import ResultStatus, VerifyCommandResult


class VerificationStatus(str, enum.Enum):
    icon: str
    is_success: bool

    @property
    def is_failed(self) -> bool:
        return not self.is_success

    def __new__(
        cls, title: str, icon: str = "", is_success: bool = False
    ) -> "VerificationStatus":

        obj = str.__new__(cls, title)
        obj._value_ = title

        obj.icon = icon
        obj.is_success = is_success
        return obj

    LIBRARY_ALL_AC = ("LIBRARY_ALL_AC", ":heavy_check_mark:", True)
    LIBRARY_PARTIAL_AC = ("LIBRARY_PARTIAL_AC", ":question:", True)
    LIBRARY_SOME_WA = ("LIBRARY_SOME_WA", ":question:", False)
    LIBRARY_ALL_WA = ("LIBRARY_ALL_WA", ":x:", False)
    LIBRARY_NO_TESTS = ("LIBRARY_NO_TESTS", ":warning:", True)
    TEST_ACCEPTED = ("TEST_ACCEPTED", ":heavy_check_mark:", True)
    TEST_WRONG_ANSWER = ("TEST_WRONG_ANSWER", ":x:", False)
    TEST_WAITING_JUDGE = ("TEST_WAITING_JUDGE", ":warning:", True)


class _VerificationStatusFlag(enum.Flag):
    TEST_NOTHING = 0
    IS_LIBRARY = enum.auto()
    HAVE_AC = enum.auto()
    HAVE_WA = enum.auto()
    HAVE_SKIP = enum.auto()

    LIBRARY_AC_WA_SKIP = IS_LIBRARY | HAVE_AC | HAVE_WA | HAVE_SKIP
    LIBRARY_AC_WA = IS_LIBRARY | HAVE_AC | HAVE_WA
    LIBRARY_AC_SKIP = IS_LIBRARY | HAVE_AC | HAVE_SKIP
    LIBRARY_AC = IS_LIBRARY | HAVE_AC
    LIBRARY_WA_SKIP = IS_LIBRARY | HAVE_WA | HAVE_SKIP
    LIBRARY_WA = IS_LIBRARY | HAVE_WA
    LIBRARY_SKIP = IS_LIBRARY | HAVE_SKIP
    LIBRARY_NOTHING = IS_LIBRARY

    TEST_AC_WA_SKIP = HAVE_AC | HAVE_WA | HAVE_SKIP
    TEST_AC_WA = HAVE_AC | HAVE_WA
    TEST_AC_SKIP = HAVE_AC | HAVE_SKIP
    TEST_AC = HAVE_AC
    TEST_WA_SKIP = HAVE_WA | HAVE_SKIP
    TEST_WA = HAVE_WA
    TEST_SKIP = HAVE_SKIP

    @classmethod
    @property
    def _conv_dict(cls) -> dict["_VerificationStatusFlag", VerificationStatus]:
        try:
            d: dict["_VerificationStatusFlag", VerificationStatus] = cls._conv_dict_attr  # type: ignore
        except AttributeError:
            d = {
                cls.LIBRARY_AC_WA_SKIP: VerificationStatus.LIBRARY_SOME_WA,
                cls.LIBRARY_AC_WA: VerificationStatus.LIBRARY_SOME_WA,
                cls.LIBRARY_AC_SKIP: VerificationStatus.LIBRARY_PARTIAL_AC,
                cls.LIBRARY_AC: VerificationStatus.LIBRARY_ALL_AC,
                cls.LIBRARY_WA_SKIP: VerificationStatus.LIBRARY_ALL_WA,
                cls.LIBRARY_WA: VerificationStatus.LIBRARY_ALL_WA,
                cls.LIBRARY_SKIP: VerificationStatus.LIBRARY_NO_TESTS,
                cls.LIBRARY_NOTHING: VerificationStatus.LIBRARY_NO_TESTS,
                cls.TEST_AC_WA_SKIP: VerificationStatus.TEST_WRONG_ANSWER,
                cls.TEST_AC_WA: VerificationStatus.TEST_WRONG_ANSWER,
                cls.TEST_AC_SKIP: VerificationStatus.TEST_WAITING_JUDGE,
                cls.TEST_AC: VerificationStatus.TEST_ACCEPTED,
                cls.TEST_WA_SKIP: VerificationStatus.TEST_WRONG_ANSWER,
                cls.TEST_WA: VerificationStatus.TEST_WRONG_ANSWER,
                cls.TEST_SKIP: VerificationStatus.TEST_WAITING_JUDGE,
                cls.TEST_NOTHING: VerificationStatus.TEST_WAITING_JUDGE,
            }
        return d

    def to_status(self) -> VerificationStatus:
        return self._conv_dict[self]


class SourceCodeStat(BaseModel):
    path: pathlib.Path
    file_input: VerificationFile
    is_verification: bool
    verification_status: VerificationStatus
    timestamp: datetime.datetime
    depends_on: set[pathlib.Path]
    required_by: set[pathlib.Path]
    verified_with: set[pathlib.Path]


def filter_edge(
    edges: dict[pathlib.Path, set[pathlib.Path]],
    excluded: set[pathlib.Path],
) -> dict[pathlib.Path, set[pathlib.Path]]:
    d: dict[pathlib.Path, set[pathlib.Path]] = {}
    for p, s in edges.items():
        if p not in excluded:
            d[p] = s - excluded
    return d


def resolve_dependency(
    *,
    input: VerificationInput,
    result: VerifyCommandResult,
    included_files: set[pathlib.Path],
) -> dict[pathlib.Path, SourceCodeStat]:
    d: dict[pathlib.Path, SourceCodeStat] = {}
    statuses: dict[pathlib.Path, _VerificationStatusFlag] = {
        p: _VerificationStatusFlag.TEST_NOTHING for p in input.files.keys()
    }

    for p, r in result.files.items():
        if p not in included_files:
            continue
        st = _VerificationStatusFlag.TEST_NOTHING
        for v in r.verifications:
            if v.status == ResultStatus.SUCCESS:
                st |= _VerificationStatusFlag.HAVE_AC
            elif v.status == ResultStatus.FAILURE:
                st |= _VerificationStatusFlag.HAVE_WA
            elif v.status == ResultStatus.SKIPPED:
                st |= _VerificationStatusFlag.HAVE_SKIP
        statuses[p] = st

    for group in input.scc():
        group &= included_files
        if not group:
            continue

        group_status = _VerificationStatusFlag.TEST_NOTHING

        for path in group:
            assert path in statuses
            group_status |= statuses[path]

        for path in group:
            depends_on = input.depends_on[path] & included_files
            required_by = input.required_by[path] & included_files
            verified_with = input.verified_with[path] & included_files

            for dep in depends_on:
                statuses[dep] |= group_status

            timestamp = git.get_commit_time(input.transitive_depends_on[path])
            file_input = input.files[path]
            is_verification = file_input.is_verification()

            flag_status = group_status
            if not is_verification:
                flag_status |= _VerificationStatusFlag.IS_LIBRARY

            d[path] = SourceCodeStat(
                path=path,
                file_input=file_input,
                is_verification=is_verification,
                depends_on=depends_on,
                required_by=required_by,
                verified_with=verified_with,
                timestamp=timestamp,
                verification_status=flag_status.to_status(),
            )
    return d
